from AnyQt.QtCore import Qt

from Orange.base import Learner
from Orange.data import Table
from Orange.i18n_config import *
from Orange.modelling import SklAdaBoostLearner, SklTreeLearner
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Msg, Input


def __(key):
    return i18n.t("widget.model.model.owadaboost." + key)


class OWAdaBoost(OWBaseLearner):
    name = __("name")
    description = __("desc")
    icon = "icons/AdaBoost.svg"
    replaces = [
        "Orange.widgets.classify.owadaboost.OWAdaBoostClassification",
        "Orange.widgets.regression.owadaboostregression.OWAdaBoostRegression",
    ]
    priority = 80
    keywords = "adaboost, boost"

    LEARNER = SklAdaBoostLearner

    class Inputs(OWBaseLearner.Inputs):
        learner = Input("Learner", Learner)

    #: Algorithms for classification problems
    algorithms = [__("gbox.samme"), __("gbox.samme_r")]
    #: Losses for regression problems
    # losses = [i18n.t("common.general.linear"), __("gbox.square"), __("gbox.exponential")]
    #
    # algorithms = ["SAMME", "SAMME.R"]
    #: Losses for regression problems
    losses = ["Linear", "Square", "Exponential"]

    n_estimators = Setting(50)
    learning_rate = Setting(1.)
    algorithm_index = Setting(1)
    loss_index = Setting(0)
    use_random_seed = Setting(False)
    random_seed = Setting(0)

    DEFAULT_BASE_ESTIMATOR = SklTreeLearner()

    class Error(OWBaseLearner.Error):
        no_weight_support = Msg(__("msg_base_learner_unsupported_weight"))

    def add_main_layout(self):
        # this is part of init, pylint: disable=attribute-defined-outside-init
        box = gui.widgetBox(self.controlArea, i18n.t("common.general.parameter"))
        self.base_estimator = self.DEFAULT_BASE_ESTIMATOR
        self.base_label = gui.label(
            box, self, __("row.base_estimator") + self.base_estimator.name.title())

        self.n_estimators_spin = gui.spin(
            box, self, "n_estimators", 1, 10000, label=__("row.num_estimator"),
            alignment=Qt.AlignRight, controlWidth=80,
            callback=self.settings_changed)
        self.learning_rate_spin = gui.doubleSpin(
            box, self, "learning_rate", 1e-5, 1.0, 1e-5,
            label=__("row.learn_rate"), decimals=5, alignment=Qt.AlignRight,
            controlWidth=80, callback=self.settings_changed)
        self.random_seed_spin = gui.spin(
            box, self, "random_seed", 0, 2 ** 31 - 1, controlWidth=80,
            label=__("checkbox_random_generator"), alignment=Qt.AlignRight,
            callback=self.settings_changed, checked="use_random_seed",
            checkCallback=self.settings_changed)

        # Algorithms
        box = gui.widgetBox(self.controlArea, __("box_boosting_method"))
        self.cls_algorithm_combo = gui.comboBox(
            box, self, "algorithm_index", label=__("row.class_algorithm"),
            items=self.algorithms,
            orientation=Qt.Horizontal, callback=self.settings_changed)
        self.reg_algorithm_combo = gui.comboBox(
            box, self, "loss_index", label=__("row.regress_loss_function"),
            items=[i18n.t("common.general.linear"), __("gbox.square"), __("gbox.exponential")],
            orientation=Qt.Horizontal, callback=self.settings_changed)

    def create_learner(self):
        if self.base_estimator is None:
            return None
        return self.LEARNER(
            estimator=self.base_estimator,
            n_estimators=self.n_estimators,
            learning_rate=self.learning_rate,
            random_state=self.random_seed,
            preprocessors=self.preprocessors,
            algorithm=self.algorithms[self.algorithm_index],
            loss=self.losses[self.loss_index].lower())

    @Inputs.learner
    def set_base_learner(self, learner):
        # base_estimator is defined in add_main_layout
        # pylint: disable=attribute-defined-outside-init
        self.Error.no_weight_support.clear()
        if learner and not learner.supports_weights:
            # Clear the error and reset to default base learner
            self.Error.no_weight_support()
            self.base_estimator = None
            self.base_label.setText(__("label.base_estimator_invalid"))
        else:
            self.base_estimator = learner or self.DEFAULT_BASE_ESTIMATOR
            self.base_label.setText(
                "Base estimator: %s" % self.base_estimator.name.title())
        self.learner = self.model = None

    def get_learner_parameters(self):
        return ((__("row.base_estimator"), self.base_estimator),
                (__("row.num_estimator"), self.n_estimators),
                (__("row.algorithm"), self.algorithms[self.algorithm_index].capitalize()),
                (__("row.loss"), self.losses[self.loss_index].capitalize()))


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWAdaBoost).run(Table("iris"))



"""
Orange Canvas Configuration

"""
import pkgutil

import random
import uuid
import warnings
import os
import sys

from typing import Dict, Any, Optional, Iterable, List

import packaging.version
import requests

from AnyQt.QtGui import (
    QPainter, QFont, QFontMetrics, QColor, QImage, QPixmap, QIcon,
    QGuiApplication
)
from AnyQt.QtCore import Qt, QPoint, QRect, QSettings

from orangecanvas import config as occonfig
from orangecanvas.config import entry_points, EntryPoint
from orangecanvas.utils.settings import config_slot
from orangewidget.workflow import config
from orangewidget.settings import set_widget_settings_dir_components

import Orange
from Orange.misc import environ

from Orange.i18n_config import *

# i18n.set('locale', 'zh')
def __(key):
    return i18n.t("orange." + key)


# generated from biolab/orange3-addons repository
OFFICIAL_ADDON_LIST = "https://bw.dashenglab.com/addons/blauwal3"

WIDGETS_ENTRY = "orange.widgets"

spec = [
    ("startup/check-updates", bool, True, "Check for updates"),

    ("startup/launch-count", int, 0, ""),

    ("reporting/machine-id", str, str(uuid.uuid4()), ""),

    ("reporting/send-statistics", bool, False, ""),

    ("reporting/permission-requested", bool, False, ""),

    ("notifications/check-notifications", bool, True, "Check for notifications"),

    ("notifications/announcements", bool, True,
     "Show notifications about Biolab announcements"),

    ("notifications/blog", bool, True,
     "Show notifications about blog posts"),

    ("notifications/new-features", bool, True,
     "Show notifications about new features"),

    ("notifications/displayed", str, 'set()',
     "Serialized set of notification IDs which have already been displayed")
]

spec = [config_slot(*t) for t in spec]


def _pixmap_from_pkg_data(package, path, format):
    contents = pkgutil.get_data(package, path)
    return QPixmap.fromImage(QImage.fromData(contents, format))


class Config(config.Config):
    """
    Orange application configuration
    """
    OrganizationDomain = "biolab.si"
    ApplicationName = __("name.application_name")
    ApplicationVersion = Orange.__version__
    AppUserModelID = "Biolab.Orange"  # AppUserModelID for windows task bar

    def init(self):
        super().init()
        self.update_language_settings()  # 添加这行代码
        QGuiApplication.setApplicationDisplayName(self.ApplicationName)
        widget_settings_dir_cfg = environ.get_path("widget_settings_dir", "")
        if widget_settings_dir_cfg:
            # widget_settings_dir is configured via config file
            set_widget_settings_dir_components(
                widget_settings_dir_cfg, self.ApplicationVersion
            )

        canvas_settings_dir_cfg = environ.get_path("canvas_settings_dir", "")
        if canvas_settings_dir_cfg:
            # canvas_settings_dir is configured via config file
            QSettings.setPath(
                QSettings.IniFormat, QSettings.UserScope,
                canvas_settings_dir_cfg
            )

        for t in spec:
            occonfig.register_setting(*t)

    def update_language_settings(self):
        """
        Update the application language settings based on user preferences.
        """
        settings = QSettings()
        language = settings.value("application-style/languages", "en")
        i18n.set('locale', language)
        self.ApplicationName = __("name.application_name")

    @staticmethod
    def application_icon():
        """
        Return the main application icon.
        """
        return QIcon(
            _pixmap_from_pkg_data(__package__, "icons/orange.ico", "ico")
        )

    @staticmethod
    def splash_screen():
        splash_n = random.randint(1, 3)
        pm = _pixmap_from_pkg_data(
            # __name__, f"icons/orange-splash-screen-{splash_n:02}.png", "png"
            __name__, f"icons/orange-splash-screen.png", "png"
        )

        version = Config.ApplicationVersion
        if version:
            version_parsed = packaging.version.Version(version)
            version_comp = version_parsed.release
            version = ".".join(map(str, version_comp[:2]))
        size = 21 if len(version) < 5 else 16
        font = QFont("Helvetica")
        font.setPixelSize(size)
        font.setBold(True)
        font.setItalic(True)
        font.setLetterSpacing(QFont.AbsoluteSpacing, 2)
        metrics = QFontMetrics(font)
        br = metrics.boundingRect(version).adjusted(-5, 0, 5, 0)
        br.moveTopLeft(QPoint(410, 218))

        p = QPainter(pm)
        p.setRenderHint(QPainter.Antialiasing)
        p.setRenderHint(QPainter.TextAntialiasing)
        p.setFont(font)
        p.setPen(QColor("#000000"))
        p.drawText(br, Qt.AlignLeft, version)
        p.end()
        return pm, QRect(88, 193, 200, 20)

    @staticmethod
    def widgets_entry_points():
        """
        Return an `EntryPoint` iterator for all 'orange.widget' entry
        points.
        """
        # Ensure the 'this' distribution's ep is the first. iter_entry_points
        # yields them in unspecified order.
        all_eps = sorted(
            entry_points(group=WIDGETS_ENTRY),
            key=lambda ep:
            0 if ep.dist.name.lower() == "Blauwal3" else 1
        )
        return iter(all_eps)

    @staticmethod
    def addon_entry_points():
        return Config.widgets_entry_points()

    @staticmethod
    def addon_defaults_list(session=None):
        # type: (Optional[requests.Session]) -> List[Dict[str, Any]]
        """
        Return a list of available add-ons.
        """
        if session is None:
            session = requests.Session()
        return session.get(OFFICIAL_ADDON_LIST).json()
    
        # import json
        # with open("/packages", "r", encoding="utf-8") as e:
        #     data = e.read()
        # json_data = json.loads(data)
        # return json_data

    @staticmethod
    def core_packages():
        # type: () -> List[str]
        """
        Return a list of 'core packages'

        These packages constitute required the application framework. They
        cannot be removes via the 'Add-on/plugins' manager. They however can
        be updated. The package that defines the application's `main()` entry
        point must always be in this list.
        """
        return ["Blauwal3 >=3.38,<4.5a"]

    @staticmethod
    def examples_entry_points():
        # type: () -> Iterable[EntryPoint]
        """
        Return an iterator over the entry points yielding 'Example Workflows'
        """
        # `iter_entry_points` yields them in unspecified order, so we order
        # them by name. The default is at the beginning, unless another
        # entrypoint precedes it alphabetically (e.g. starting with '!').
        default_ep = EntryPoint(
            "Blauwal3", "Orange.canvas.workflows", "orange.widgets.tutorials",
        )

        all_ep = list(entry_points(group="orange.widgets.tutorials"))
        all_ep.append(default_ep)
        all_ep.sort(key=lambda x: x.name)
        return iter(all_ep)

    APPLICATION_URLS = {
        #: Submit a bug report action in the Help menu
        "Bug Report": "https://github.com/biolab/orange3/issues",
        #: A url quick tour/getting started url
        "Quick Start": "https://bw.dashenglab.com/5fbb9705a6e5e67080638ac6.html",
        #: The 'full' documentation, should be something like current /docs/
        #: but specific for 'Visual Programing' only
        "Documentation": "https://bw.dashenglab.com/5f96afc397e57e5494285683.html",
        #: YouTube tutorials
        "Screencasts":
            "https://www.youtube.com/watch"
            "?v=HXjnDIgGDuI&list=PLmNPvQr9Tf-ZSDLwOzxpvY-HrE0yv-8Fy&index=1",
        #: Used for 'Submit Feedback' action in the help menu
        "Feedback": "https://orange.biolab.si/survey/long.html",
    }


def init():
    # left for backwards compatibility
    raise RuntimeError("This is not the init you are looking for.")


def data_dir():
    """
    Return the Orange application data directory. If the directory path
    does not yet exists then create it.
    """
    path = os.path.join(environ.data_dir(), "canvas")
    try:
        os.makedirs(path, exist_ok=True)
    except OSError:
        pass
    return path


def cache_dir():
    """
    Return the Orange application cache directory. If the directory path
    does not yet exists then create it.
    """
    path = os.path.join(environ.cache_dir(), "canvas")
    try:
        os.makedirs(path, exist_ok=True)
    except OSError:
        pass
    return path


def log_dir():
    """
    Return the application log directory.
    """
    if sys.platform == "darwin":
        name = Config.ApplicationName
        logdir = os.path.join(os.path.expanduser("~/Library/Logs"), name)
    else:
        logdir = data_dir()

    logdir = environ.get_path("log_dir", logdir)

    try:
        os.makedirs(logdir, exist_ok=True)
    except OSError:
        pass
    return logdir


def widget_settings_dir():
    """
    Return the widget settings directory.

    .. deprecated:: 3.38
    """
    warnings.warn(
        __("tip.deprecate").format(__name__),
        DeprecationWarning, stacklevel=2
    )
    import orangewidget.settings
    return orangewidget.settings.widget_settings_dir()


def widgets_entry_points():
    return Config.widgets_entry_points()


def addon_entry_points():
    return Config.addon_entry_points()


def splash_screen():
    return Config.splash_screen()


def application_icon():
    return Config.application_icon()



import threading
import textwrap

import numpy as np

from Orange.widgets import widget, gui
from Orange.widgets.utils.localization import pl
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input
from Orange.widgets import report
from Orange.i18n_config import *
from Orange.data import \
    Table, StringVariable, DiscreteVariable, ContinuousVariable

try:
    from Orange.data.sql.table import SqlTable
except ImportError:
    def is_sql(_):
        return False
else:
    def is_sql(data):
        return isinstance(data, SqlTable)


def __(key):
    return i18n.t("widget.data.data.owdatainfo." + key)


class OWDataInfo(widget.OWWidget):
    name = __("name")
    id = "orange.widgets.data.info"
    description = __("desc")
    icon = "icons/DataInfo.svg"
    priority = 80
    category = "Data"
    keywords = "data info, information, inspect"

    class Inputs:
        data = Input("Data", Table, label=i18n.t("widget.data.data.common.data"))

    want_main_area = False
    buttons_area_orientation = None
    resizing_enabled = False

    def __init__(self):
        super().__init__()

        self.data_desc = {}
        self.data_attrs = {}
        self.description = gui.widgetLabel(
            gui.vBox(self.controlArea, box=__("box.set_name")))
        self.attributes = gui.widgetLabel(
            gui.vBox(self.controlArea, box=__("box.data_attr")))

    @Inputs.data
    def data(self, data):
        if data is None:
            self.data_desc = self.data_attrs = {}
            self.update_info()
        else:
            self.data_desc = {
                label: value
                for label, func in ((__("row.num_var"), self._p_name),
                                    (__("row.location"), self._p_location),
                                    (__("row.size"), self._p_size),
                                    (__("row.features"), self._p_features),
                                    (__("row.targets"), self._p_targets),
                                    (__("row.metas"), self._p_metas),
                                    (__("row.missing_data"), self._p_missing))
                if bool(value := func(data))}
            self.data_attrs = data.attributes
            self.update_info()

            if is_sql(data):
                def set_exact_length():
                    self.data_desc[__("row.size")] = self._p_size(data, exact=True)
                    self.update_info()

                threading.Thread(target=set_exact_length).start()

    def update_info(self):
        style = """<style>
                       th { text-align: right; vertical-align: top; }
                       th, td { padding-top: 4px; line-height: 125%}
                    </style>"""

        def dict_as_table(d):
            return "<table>" + \
                   "".join(f"<tr><th>{label}: </th><td>" + \
                           '<br/>'.join(textwrap.wrap(value, width=60)) + \
                           "</td></tr>"
                           for label, value in d.items()) + \
                   "</table>"

        if not self.data_desc:
            self.description.setText(__("msg.no_data"))
        else:
            self.description.setText(style + dict_as_table(self.data_desc))
        self.attributes.setHidden(not self.data_attrs)
        if self.data_attrs:
            self.attributes.setText(
                style + dict_as_table({k: str(v)
                                       for k, v in self.data_attrs.items()}))

    def send_report(self):
        if self.data_desc:
            self.report_items(__("report.data_table_properties"), self.data_desc)
        if self.data_attrs:
            self.report_items(__("report.additional_attributes"), self.data_attrs)

    @staticmethod
    def _p_name(data):
        return getattr(data, "name", "-")

    @staticmethod
    def _p_location(data):
        if not is_sql(data):
            return None

        connection_string = ' '.join(
            f'{key}={value}'
            for key, value in data.connection_params.items()
            if value is not None and key != 'password')
        return f"SQL Table using connection:<br/>{connection_string}"

    @staticmethod
    def _p_size(data, exact=False):
        exact = exact or is_sql(data)
        if exact:
            n = len(data)
            desc = f"{n} {pl(n, 'row')}"
        else:
            n = data.approx_len()
            desc = f"~{n} {pl(n, 'row')}"
        ncols = len(data.domain.variables) + len(data.domain.metas)
        desc += f", {ncols} {pl(ncols, 'column')}"

        sparseness = [s for s, m in (("features", data.X_density),
                                     ("meta attributes", data.metas_density),
                                     ("targets", data.Y_density)) if m() > 1]
        if sparseness:
            desc += "; sparse {', '.join(sparseness)}"
        return desc

    @classmethod
    def _p_features(cls, data):
        return cls._pack_var_counts(data.domain.attributes)

    def _p_targets(self, data):
        if class_var := data.domain.class_var:
            if class_var.is_continuous:
                self.targets = __("row.num_var")
            else:
                self.targets = __("row.categorical_value") \
                    .format(len(class_var.values))
        if class_vars := data.domain.class_vars:
            disc_class = self._count(class_vars, DiscreteVariable)
            cont_class = self._count(class_vars, ContinuousVariable)
            if not cont_class:
                return f"{disc_class} categorical {pl(disc_class, 'target')}"
            elif not disc_class:
                return f"{cont_class} numeric {pl(cont_class, 'target')}"
            return "multi-target data,<br/>" + self._pack_var_counts(class_vars)

    @classmethod
    def _p_metas(cls, data):
        return cls._pack_var_counts(data.domain.metas)

    @staticmethod
    def _p_missing(data: Table):
        if is_sql(data):
            return "(not checked for SQL data)"

        counts = []
        for name, part, n_miss in ((pl(len(data.domain.attributes), "feature"),
                                    data.X, data.get_nan_count_attribute()),
                                   (pl(len(data.domain.class_vars), "targets"),
                                    data.Y, data.get_nan_count_class()),
                                   (pl(len(data.domain.metas), "meta variable"),
                                    data.metas, data.get_nan_count_metas())):
            if n_miss:
                counts.append(
                    f"{n_miss} ({n_miss / np.prod(part.shape):.1%}) in {name}")
        if not counts:
            return "none"
        return ", ".join(counts)

    @staticmethod
    def _count(s, tpe):
        return sum(isinstance(x, tpe) for x in s)

    @classmethod
    def _pack_var_counts(cls, s):
        counts = (
            (name, cls._count(s, type_))
            for name, type_ in (("categorical", DiscreteVariable),
                                ("numeric", ContinuousVariable),
                                ("text", StringVariable)))
        return ", ".join(f"{count} {name}" for name, count in counts if count)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWDataInfo).run(Table("heart_disease"))



import os
import unittest
from unittest.mock import patch

import numpy as np
from AnyQt.QtCore import QMimeData, QUrl

from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import dragDrop
from Orange.widgets.unsupervised.owdistancefile \
    import OWDistanceFile, OWDistanceFileDropHandler

import Orange.tests


class TestOWDistanceFile(WidgetTest):
    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(OWDistanceFile)

    def open_file(self, filename):
        filename = os.path.join(os.path.split(Orange.tests.__file__)[0],
                                filename)
        self.widget.add_path(filename)
        self.widget.open_file()

    def test_non_square(self):
        self.open_file("xlsx_files/distances_nonsquare.xlsx")
        self.assertIsNone(self.get_output(self.widget.Outputs.distances))
        self.assertTrue(self.widget.Error.non_square_matrix.is_shown())
        self.open_file("xlsx_files/distances_with_nans.xlsx")
        self.assertFalse(self.widget.Error.non_square_matrix.is_shown())

    def test_nan_to_num(self):
        self.open_file("xlsx_files/distances_with_nans.xlsx")
        dist = self.get_output(self.widget.Outputs.distances)
        np.testing.assert_equal(dist, [[1, 2, 3], [4, 5, 0], [7, 0, 9]])

    def test_drop_file(self):
        mime = QMimeData()
        mime.setUrls([QUrl("https://example.com/a.html")])
        with patch.object(self.widget, "open_file") as r:
            self.assertFalse(dragDrop(self.widget, mime))
            r.assert_not_called()

        mime.setUrls([QUrl.fromLocalFile("file.notsupported")])
        with patch.object(self.widget, "open_file") as r:
            self.assertFalse(dragDrop(self.widget, mime))
            r.assert_not_called()

        filename = os.path.normpath(os.path.join(__file__, "../test.dst"))
        mime.setUrls([QUrl.fromLocalFile(filename)])
        with patch.object(self.widget, "open_file") as r:
            self.assertTrue(dragDrop(self.widget, mime))
            self.assertEqual(os.path.normpath(self.widget.last_path()), filename)
            r.assert_called()


class TestOWDistanceFileDropHandler(unittest.TestCase):
    def test_canDropFile(self):
        handler = OWDistanceFileDropHandler()
        self.assertTrue(handler.canDropFile("test.dst"))
        self.assertTrue(handler.canDropFile("test.xlsx"))
        self.assertFalse(handler.canDropFile("test.bin"))

    def test_parametersFromFile(self):
        handler = OWDistanceFileDropHandler()
        r = handler.parametersFromFile("test.dst")
        self.assertEqual(r["recent_paths"][0].basename, "test.dst")



"""Various small utilities that might be useful everywhere"""
import logging
import os
import time
import inspect
import datetime
import math
import functools
import importlib.resources
from contextlib import contextmanager
from importlib.metadata import distribution
from typing import TYPE_CHECKING, Callable, Union, Optional
from weakref import WeakKeyDictionary
from enum import Enum as _Enum
from functools import wraps, partial
from operator import attrgetter
from itertools import chain, count, repeat

from collections import namedtuple
import warnings

# Exposed here for convenience. Prefer patching to try-finally blocks
from unittest.mock import patch  # pylint: disable=unused-import

import numpy as np

# Backwards-compat
from Orange.data.util import scale  # pylint: disable=unused-import

if TYPE_CHECKING:
    from numpy.typing import DTypeLike


log = logging.getLogger(__name__)


class OrangeWarning(UserWarning):
    pass


class OrangeDeprecationWarning(OrangeWarning, DeprecationWarning):
    pass


warnings.simplefilter('default', OrangeWarning)

if os.environ.get('ORANGE_DEPRECATIONS_ERROR'):
    warnings.simplefilter('error', OrangeDeprecationWarning)


def _log_warning(msg):
    """
    Replacement for `warnings._showwarnmsg_impl` that logs the warning
    Logs the warning in the appropriate list, or passes it to the original
    function if the warning wasn't issued within the log_warnings context.
    """
    for frame in inspect.stack():
        if frame.frame in warning_loggers:
            warning_loggers[frame.frame].append(msg)
            break
    else:
        __orig_showwarnmsg_impl(msg)


@contextmanager
def log_warnings():
    """
    logs all warnings that occur within context, including warnings from calls.

    ```python
    with log_warnings() as warnings:
       ...
    ```

    Unlike `warnings.catch_warnings(record=True)`, this manager is thread-safe
    and will only log warning from this thread. It does so by storing the
    stack frame within which the context is created, and then checking the
    stack when the warning is issued.

    Nesting of `log_warnings` within the same function will raise an error.
    If `log_wanings` are nested within function calls, the warning is logged
    in the inner-most context.

    If `catch_warnings` is used within the `log_warnings` context, logging is
    disabled until the `catch_warnings` exits. This looks inevitable (without
    patching `catch_warnings`, which I'd prefer not to do).

    If `catch_warnings` is used outside this context, everything, including
    warning filtering, should work as expected.

    Note: the method imitates `catch_warnings` by patching the `warnings`
    module's internal function `_showwarnmsg_impl`. Python (as of version 3.9)
    doesn't seem to offer any other way of catching the warnings. This function
    was introduced in Python 3.6, so we cover all supported versions. If it is
    ever removed, unittests will crash, so we'll know. :)
    """
    # currentframe().f_back is `contextmanager`'s __enter__
    frame = inspect.currentframe().f_back.f_back
    if frame in warning_loggers:
        raise ValueError("nested log_warnings")
    try:
        warning_loggers[frame] = []
        yield warning_loggers[frame]
    finally:
        del warning_loggers[frame]


# pylint: disable=protected-access
warning_loggers = {}
__orig_showwarnmsg_impl = warnings._showwarnmsg_impl
warnings._showwarnmsg_impl = _log_warning


def resource_filename(path):
    """
    Return the resource filename path relative to the Orange package.
    """
    path = importlib.resources.files("Orange").joinpath(path)
    return str(path)


def get_entry_point(dist, group, name):
    """
    Load and return the entry point from the distribution.
    """
    dist = distribution(dist)
    eps = dist.entry_points.select(group=group, name=name)
    ep = next(iter(eps))
    return ep.load()


def deprecated(obj):
    """
    Decorator. Mark called object deprecated.

    Parameters
    ----------
    obj: callable or str
        If callable, it is marked as deprecated and its calling raises
        OrangeDeprecationWarning. If str, it is the alternative to be used
        instead of the decorated function.

    Returns
    -------
    f: wrapped callable or decorator
        Returns decorator if obj was str.

    Examples
    --------
    >>> @deprecated
    ... def old():
    ...     return 'old behavior'
    >>> old()  # doctest: +SKIP
    /... OrangeDeprecationWarning: Call to deprecated ... old ...
    'old behavior'

    >>> class C:
    ...     @deprecated('C.new()')
    ...     def old(self):
    ...         return 'old behavior'
    ...     def new(self):
    ...         return 'new behavior'
    >>> C().old() # doctest: +SKIP
    /... OrangeDeprecationWarning: Call to deprecated ... C.old ...
      use use C.new() instead ...
    'old behavior'
    """
    alternative = f'; use {obj} instead' if isinstance(obj, str) else ''

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            name = func.__name__
            if hasattr(func, "__self__"):
                name = f'{func.__self__.__class__}.{name}'
            warnings.warn(f'Call to deprecated {name}{alternative}',
                          OrangeDeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper

    return decorator if alternative else decorator(obj)


# This should look like decorator, not a class, pylint: disable=invalid-name
class allot:
    """
    Decorator that allows a function only a specified portion of time per call.

    Usage:

    ```
    @allot(0.2, overflow=of)
    def f(x):
       ...
    ```

    The above function is allotted 0.2 second per second. If it runs for 0.2 s,
    all subsequent calls in the next second (after the start of the call) are
    ignored. If it runs for 0.1 s, subsequent calls in the next 0.5 s are
    ignored. If it runs for a second, subsequent calls are ignored for 5 s.

    An optional overflow function can be given as a keyword argument
    `overflow`. This function must have the same signature as the wrapped
    function and is called instead of the original when the call is blocked.

    If the overflow function is not given, the wrapped function must not return
    result. This is because without the overflow function, the wrapper has no
    value to return when the call is skipped.

    The decorator adds a method `call` to force the call, e.g. by calling
    f.call(5), in the above case. The used up time still counts for the
    following (non-forced) calls.

    The decorator also adds two attributes:

    - f.last_call_duration is the duration of the last call (in seconds)
    - f.no_call_before contains the time stamp when the next call will be made.

    The decorator can be used for functions and for methods.

    A non-parametrized decorator doesn't block any calls and only adds
    last_call_duration, so that it can be used for timing.
    """

    try:
        __timer = time.thread_time
    except AttributeError:
        # thread_time is not available on macOS
        __timer = time.process_time

    def __new__(cls: type, arg: Union[None, float, Callable], *,
                overflow: Optional[Callable] = None,
                _bound_methods: Optional[WeakKeyDictionary] = None):
        self = super().__new__(cls)

        if arg is None or isinstance(arg, float):
            # Parametrized decorator
            if arg is not None:
                assert arg > 0

            def set_func(func):
                self.__init__(func,
                              overflow=overflow,
                              _bound_methods=_bound_methods)
                self.allotted_time = arg
                return self

            return set_func

        else:
            # Non-parametrized decorator
            self.allotted_time = None
            return self

    def __init__(self,
                 func: Callable, *,
                 overflow: Optional[Callable] = None,
                 _bound_methods: Optional[WeakKeyDictionary] = None):
        assert callable(func)
        self.func = func
        self.overflow = overflow
        functools.update_wrapper(self, func)

        self.no_call_before = 0
        self.last_call_duration = None

        # Used by __get__; see a comment there
        if _bound_methods is None:
            self.__bound_methods = WeakKeyDictionary()
        else:
            self.__bound_methods = _bound_methods

    # If we are wrapping a method, __get__ is called to bind it.
    # Create a wrapper for each instance and store it, so that each instance's
    # method gets its share of time.
    def __get__(self, inst, cls):
        if inst is None:
            return self

        if inst not in self.__bound_methods:
            # __bound_methods caches bound methods per instance. This is not
            # done for perfoamnce. Bound methods can be rebound, even to
            # different instances or even classes, e.g.
            # >>> x = f.__get__(a, A)
            # >>> y = x.__get__(b, B)
            # >>> z = x.__get__(a, A)
            # After this, we want `x is z`, there shared caching. This looks
            # bizarre, but let's keep it safe. At least binding to the same
            # instance, f.__get__(a, A),__get__(a, A), sounds reasonably
            # possible.
            cls = type(self)
            bound_overflow = self.overflow and self.overflow.__get__(inst, cls)
            decorator = cls(
                self.allotted_time,
                overflow=bound_overflow,
                _bound_methods=self.__bound_methods)
            self.__bound_methods[inst] = decorator(self.func.__get__(inst, cls))

        return self.__bound_methods[inst]

    def __call__(self, *args, **kwargs):
        if self.__timer() < self.no_call_before:
            if self.overflow is None:
                return None
            return self.overflow(*args, **kwargs)
        return self.call(*args, **kwargs)

    def call(self, *args, **kwargs):
        start = self.__timer()
        result = self.func(*args, **kwargs)
        self.last_call_duration = self.__timer() - start
        if self.allotted_time is not None:
            if self.overflow is None:
                assert result is None, "skippable function cannot return a result"
            self.no_call_before = start + self.last_call_duration / self.allotted_time
        return result


def literal_eval(literal):
    import ast  # pylint: disable=import-outside-toplevel
    # ast.literal_eval does not parse empty set ¯\_(ツ)_/¯

    if literal == "set()":
        return set()
    return ast.literal_eval(literal)


op_map = {
    '==': lambda a, b: a == b,
    '>=': lambda a, b: a >= b,
    '<=': lambda a, b: a <= b,
    '>': lambda a, b: a > b,
    '<': lambda a, b: a < b
}


_Requirement = namedtuple("_Requirement", ["name", "op", "value"])


bool_map = {
    "True": True,
    "true": True,
    1: True,
    "False": False,
    "false": False,
    0: False
}


def requirementsSatisfied(required_state, local_state, req_type=None):
    """
    Checks a list of requirements against a dictionary representing local state.

    Args:
        required_state ([str]): List of strings representing required state
                                using comparison operators
        local_state (dict): Dictionary representing current state
        req_type (type): Casts values to req_type before comparing them.
                         Defaults to local_state type.
    """
    for req_string in required_state:
        # parse requirement
        req = None
        for op_str, op in op_map.items():
            split = req_string.split(op_str)
            # if operation is not in req_string, continue
            if len(split) == 2:
                req = _Requirement(split[0], op, split[1])
                break

        if req is None:
            log.error("Invalid requirement specification: %s", req_string)
            return False

        compare_type = req_type or type(local_state[req.name])
        # check if local state satisfies required state (specification)
        if compare_type is bool:
            # boolean is a special case, where simply casting to bool does not produce target result
            required_value = bool_map[req.value]
        else:
            required_value = compare_type(req.value)
        local_value = compare_type(local_state[req.name])

        # finally, compare the values
        if not req.op(local_value, required_value):
            return False
    return True


def try_(func, default=None):
    """Try return the result of func, else return default."""
    try:
        return func()
    except Exception:  # pylint: disable=broad-except
        return default


def flatten(lst):
    """Flatten iterable a single level."""
    return chain.from_iterable(lst)


class Registry(type):
    """Metaclass that registers subtypes."""
    def __new__(mcs, name, bases, attrs):
        cls = type.__new__(mcs, name, bases, attrs)
        if not hasattr(cls, 'registry'):
            cls.registry = {}
        else:
            cls.registry[name] = cls
        return cls

    def __iter__(cls):
        return iter(cls.registry)

    def __str__(cls):
        if cls in cls.registry.values():
            return cls.__name__
        return f'{cls.__name__}({{{", ".join(cls.registry)}}})'


# it is what it is, we keep for compatibility:
# pylint: disable=keyword-arg-before-vararg
def namegen(prefix='_', *args, spec_count=count, **kwargs):
    """Continually generate names with `prefix`, e.g. '_1', '_2', ..."""
    # pylint: disable=stop-iteration-return
    spec_count = iter(spec_count(*args, **kwargs))
    while True:
        yield prefix + str(next(spec_count))


def export_globals(globals, module_name):
    """
    Return list of important for export globals (callables, constants) from
    `globals` dict, defined in module `module_name`.

    Usage
    -----
    In some module, on the second-to-last line:

    __all__ = export_globals(globals(), __name__)

    """
    return [getattr(v, '__name__', k)
            for k, v in globals.items()                          # export
            if ((callable(v) and v.__module__ == module_name     # callables from this module
                 or k.isupper()) and                             # or CONSTANTS
                not getattr(v, '__name__', k).startswith('_'))]  # neither marked internal


_NOTSET = object()


def deepgetattr(obj, attr, default=_NOTSET):
    """Works exactly like getattr(), except that attr can be a nested attribute
    (e.g. "attr1.attr2.attr3").
    """
    try:
        return attrgetter(attr)(obj)
    except AttributeError:
        if default is _NOTSET:
            raise
        return default


def color_to_hex(color):
    # pylint: disable=consider-using-f-string
    return "#{:02X}{:02X}{:02X}".format(*color)


def hex_to_color(s):
    return int(s[1:3], 16), int(s[3:5], 16), int(s[5:7], 16)


def inherit_docstrings(cls):
    """Inherit methods' docstrings from first superclass that defines them"""
    for method in cls.__dict__.values():
        if inspect.isfunction(method) and method.__doc__ is None:
            for parent in cls.__mro__[1:]:
                doc = getattr(parent, method.__name__, None).__doc__
                if doc:
                    method.__doc__ = doc
                    break
    return cls


class Enum(_Enum):
    """Enum that represents itself with the qualified name, e.g. Color.red"""
    __repr__ = _Enum.__str__


def interleave(seq1, seq2):
    """
    Interleave elements of `seq2` between consecutive elements of `seq1`.

    Example
    -------
    >>> list(interleave([1, 3, 5], [2, 4]))
    [1, 2, 3, 4, 5]
    >>> list(interleave([1, 2, 3, 4], repeat("<")))
    [1, '<', 2, '<', 3, '<', 4]
    """
    iterator1, iterator2 = iter(seq1), iter(seq2)
    try:
        leading = next(iterator1)
    except StopIteration:
        pass
    else:
        for element in iterator1:
            yield leading
            try:
                yield next(iterator2)
            except StopIteration:
                return
            leading = element
        yield leading


def Reprable_repr_pretty(name, itemsiter, printer, cycle):
    # type: (str, Iterable[Tuple[str, Any]], Ipython.lib.pretty.PrettyPrinter, bool) -> None
    if cycle:
        printer.text(f"{name}(...)")
    else:
        def printitem(field, value):
            printer.text(field + "=")
            printer.pretty(value)

        def printsep():
            printer.text(",")
            printer.breakable()

        itemsiter = (partial(printitem, *item) for item in itemsiter)
        sepiter = repeat(printsep)

        with printer.group(len(name) + 1, f"{name}(", ")"):
            for part in interleave(itemsiter, sepiter):
                part()
                part()


class _Undef:
    def __repr__(self):
        return "<?>"
_undef = _Undef()


class Reprable:
    """A type that inherits from this class has its __repr__ string
    auto-generated so that it "[...] should look like a valid Python
    expression that could be used to recreate an object with the same
    value [...]" (see See Also section below).

    This relies on the instances of type to have attributes that
    match the arguments of the type's constructor. Only the values that
    don't match the arguments' defaults are printed, i.e.:

        >>> class C(Reprable):
        ...     def __init__(self, a, b=2):
        ...         self.a = a
        ...         self.b = b
        >>> C(1, 2)
        C(a=1)
        >>> C(1, 3)
        C(a=1, b=3)

    If Reprable instances define `_reprable_module`, that string is used
    as a fully-qualified module name and is printed. `_reprable_module`
    can also be True in which case the type's home module is used.

        >>> class C(Reprable):
        ...     _reprable_module = True
        >>> C()
        Orange.util.C()
        >>> class C(Reprable):
        ...     _reprable_module = 'something_else'
        >>> C()
        something_else.C()
        >>> class C(Reprable):
        ...     class ModuleResolver:
        ...         def __str__(self):
        ...             return 'magic'
        ...     _reprable_module = ModuleResolver()
        >>> C()
        magic.C()

    See Also
    --------
    https://docs.python.org/3/reference/datamodel.html#object.__repr__
    """
    _reprable_module = ''

    def _reprable_fields(self):
        # type: () -> Iterable[Tuple[str, Any]]
        cls = self.__class__
        sig = inspect.signature(cls.__init__)
        for param in sig.parameters.values():
            # Skip self, *args, **kwargs
            if param.name != 'self' and \
                    param.kind not in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
                yield param.name, param.default

    # pylint: disable=unused-argument
    def _reprable_omit_param(self, name, default, value):
        if default is value:
            return True
        if type(default) is type(value):
            try:
                return default == value
            except (ValueError, TypeError):
                return False
        else:
            return False

    def _reprable_items(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error", DeprecationWarning)
            warnings.simplefilter("error", PendingDeprecationWarning)
            for name, default in self._reprable_fields():
                try:
                    value = getattr(self, name)
                except (DeprecationWarning, PendingDeprecationWarning):
                    continue
                except AttributeError:
                    value = _undef
                if not self._reprable_omit_param(name, default, value):
                    yield name, default, value

    def _repr_pretty_(self, p, cycle):
        """IPython pretty print hook."""
        module = self._reprable_module
        if module is True:
            module = self.__class__.__module__

        nameparts = (([str(module)] if module else []) +
                     [self.__class__.__name__])
        name = ".".join(nameparts)
        Reprable_repr_pretty(
            name, ((f, v) for f, _, v in self._reprable_items()),
            p, cycle)

    def __repr__(self):
        module = self._reprable_module
        if module is True:
            module = self.__class__.__module__
        nameparts = (([str(module)] if module else []) +
                     [self.__class__.__name__])
        name = ".".join(nameparts)
        items = ", ".join(f"{f}={repr(v)}"
                          for f, _, v in self._reprable_items())
        return f"{name}({items})"


def wrap_callback(progress_callback, start=0, end=1):
    """
    Wraps a progress callback function to allocate it end-start proportion
    of an execution time.

    :param progress_callback: callable
    :param start: float
    :param end: float
    :return: callable
    """
    @wraps(progress_callback)
    def func(progress, *args, **kwargs):
        adjusted_progress = start + progress * (end - start)
        return progress_callback(adjusted_progress, *args, **kwargs)
    return func


def dummy_callback(*_, **__):
    """ A dummy callable. """
    return 1


def utc_from_timestamp(timestamp) -> datetime.datetime:
    """
    Return the UTC datetime corresponding to the POSIX timestamp.
    """
    return datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc) + \
           datetime.timedelta(seconds=float(timestamp))


def frompyfunc(func: Callable, nin: int, nout: int, dtype: 'DTypeLike'):
    """
    Wrap an `func` callable into an ufunc-like function with `out`, `dtype`,
    `where`, ... parameters. The `dtype` is used as the default.

    Unlike numpy.frompyfunc this function always returns output array of
    the specified `dtype`. Note that the conversion is space efficient.
    """
    func_ = np.frompyfunc(func, nin, nout)

    @wraps(func)
    def funcv(*args, out=None, dtype=dtype, casting="unsafe", **kwargs):
        if not args:
            raise TypeError
        args = [np.asanyarray(a) for a in args]
        args = np.broadcast_arrays(*args)
        shape = args[0].shape
        have_out = out is not None
        if out is None and dtype is not None:
            out = np.empty(shape, dtype)

        res = func_(*args, out, dtype=dtype, casting=casting, **kwargs)
        if res.shape == () and not have_out:
            return res.item()
        else:
            return res

    return funcv


_isnan = math.isnan


def nan_eq(a, b) -> bool:
    """
    Same as `a == b` except where both `a` and `b` are  NaN values in which
    case `True` is returned.

    .. seealso:: nan_hash_stand
    """
    try:
        both_nan = _isnan(a) and _isnan(b)
    except TypeError:
        return a == b
    else:
        return both_nan or a == b


def nan_hash_stand(value):
    """
    If `value` is a NaN then return a singular global *standin* NaN instance,
    otherwise return `value` unchanged.

    Use this where a hash of `value` is needed and `value` might be a NaN
    to account for distinct hashes of NaN instances.

    E.g. the folowing `__eq__` and `__hash__` pairs would be ill-defined for
    `A(float("nan"))` instances if `nan_hash_stand` and `nan_eq` were not
    used.
    >>> class A:
    ...     def __init__(self, v): self.v = v
    ...     def __hash__(self): return hash(nan_hash_stand(self.v))
    ...     def __eq__(self, other): return nan_eq(self.v, other.v)
    """
    try:
        if _isnan(value):
            return math.nan
    except TypeError:
        pass
    return value


# For best result, keep this at the bottom
__all__ = export_globals(globals(), __name__)

# ONLY NON-EXPORTED VALUES BELOW HERE



from unittest.mock import Mock

import numpy as np

from Orange import classification, evaluation
from Orange.data import Table, Domain, DiscreteVariable
from Orange.evaluation import Results
from Orange.evaluation.performance_curves import Curves
from Orange.tests import test_filename

from Orange.widgets.tests.base import WidgetTest


class EvaluateTest(WidgetTest):
    def setUp(self):
        super().setUp()

        n, p = (0, 1)
        actual, probs = np.array([
            (p, .8), (n, .7), (p, .6), (p, .55), (p, .54), (n, .53), (n, .52),
            (p, .51), (n, .505), (p, .4), (n, .39), (p, .38), (n, .37),
            (n, .36), (n, .35), (p, .34), (n, .33), (p, .30), (n, .1)]).T
        self.curves = Curves(actual, probs)
        probs2 = (probs + 1) / 2
        self.curves2 = Curves(actual, probs2)
        pred = probs > 0.5
        pred2 = probs2 > 0.5
        probs = np.vstack((1 - probs, probs)).T
        probs2 = np.vstack((1 - probs2, probs2)).T
        domain = Domain([], DiscreteVariable("y", values=("a", "b")))
        self.results = Results(
            domain=domain,
            actual=actual,
            folds=np.array([Ellipsis]),
            models=np.array([[Mock(), Mock()]]),
            row_indices=np.arange(19),
            predicted=np.array((pred, pred2)),
            probabilities=np.array([probs, probs2]))

        self.lenses = data = Table(test_filename("datasets/lenses.tab"))
        majority = classification.MajorityLearner()
        majority.name = "majority"
        knn3 = classification.KNNLearner(n_neighbors=3)
        knn3.name = "knn-3"
        knn1 = classification.KNNLearner(n_neighbors=1)
        knn1.name = "knn-1"
        self.lenses_results = evaluation.TestOnTestData(
            store_data=True, store_models=True)(
                data=data[::2], test_data=data[1::2],
                learners=[majority, knn3, knn1])
        self.lenses_results.learner_names = ["majority", "knn-3", "knn-1"]

    def test_many_evaluation_results(self):
        if not hasattr(self, "widget"):
            return

        data = Table("iris")
        learners = [
            classification.MajorityLearner(),
            classification.LogisticRegressionLearner(),
            classification.TreeLearner(),
            classification.SVMLearner(),
            classification.KNNLearner(),
            classification.CN2Learner(),
            classification.SGDClassificationLearner(),
            classification.RandomForestLearner(),
            classification.NaiveBayesLearner(),
            classification.SGDClassificationLearner()
        ]
        res = evaluation.CrossValidation(k=2, store_data=True)(data, learners)
        # this is a mixin; pylint: disable=no-member
        self.send_signal("Evaluation Results", res)



import unittest
from unittest.mock import patch

import numpy as np
from scipy import sparse as sp

from Orange.data import \
    Table, Domain, ContinuousVariable, DiscreteVariable, StringVariable
from Orange.widgets.data.owdatainfo import OWDataInfo
from Orange.widgets.tests.base import WidgetTest


class TestOWDataInfo(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWDataInfo)

    def test_data(self):
        # I guess we don't want to test specific output tests, just different
        # combinations that must not crash
        a, b, c = (DiscreteVariable(n) for n in "abc")
        x, y, z = (ContinuousVariable(n) for n in "xyz")
        m, n = (StringVariable(n) for n in "mn")
        meta_s = np.array([["foo", "bar", ""]]).T
        meta_c = np.array([[3.14, np.nan, np.nan]]).T
        metadata = np.hstack((meta_s, meta_c))
        self.widget.send_report()
        for attrs, classes, metas, metad in (((a, b, c), (), (), None),
                                             ((a, b, c, x), (y,), (), None),
                                             ((a, b, c), (y, x), (m, ), meta_s),
                                             ((a, b, c), (y, ), (x, ), meta_c),
                                             ((a, b), (y, x, c), (m, ), meta_s),
                                             ((a, ), (b, c), (m, ), meta_s),
                                             ((a, b, x), (c, ), (m, y), metadata),
                                             ((), (c, ), (m, y), metadata)):
            data = Table.from_numpy(
                Domain(attrs, classes, metas),
                np.zeros((3, len(attrs))),
                np.zeros((3, len(classes))),
                metad)
            data.attributes = {"att 1": 1, "att 2": True, "att 3": 3}
            if metas:
                data.name = "name"
            self.send_signal(self.widget.Inputs.data, data)
            self.widget.send_report()
        self.send_signal(self.widget.Inputs.data, None)
        self.widget.send_report()

        data.attributes = {"foo": "bar"}
        self.send_signal(self.widget.Inputs.data, None)
        self.widget.send_report()

    def test_sparse(self):
        x, y, z, u, w = (ContinuousVariable(n) for n in "xyzuw")
        data = Table.from_numpy(
            Domain([x, y], z, [u, w]),
            sp.csc_matrix(np.random.randint(0, 1, (5, 2))),
            sp.csc_matrix(np.random.randint(0, 1, (5, 1))),
            sp.csc_matrix(np.random.randint(0, 1, (5, 2))))
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.send_report()

    def test_sql(self):
        class SqlTable(Table):
            connection_params = {"foo": "bar"}

        class Thread:
            def __init__(self, target):
                self.target = target

            def start(self):
                self.target()

        w = self.widget

        domain = Domain([ContinuousVariable("y")])

        with patch("Orange.widgets.data.owdatainfo.SqlTable", new=SqlTable), \
                patch("threading.Thread", new=Thread), \
                patch.object(self.widget, "_p_size", wraps=self.widget._p_size) as p_size:

            self.send_signal(w.Inputs.data, Table.from_numpy(domain, [[42]]))
            p_size.assert_called_once()
            p_size.reset_mock()

            d = SqlTable.from_numpy(domain, [[42]])
            self.send_signal(w.Inputs.data, d)
            self.assertEqual(p_size.call_count, 2)
            self.assertEqual(p_size.call_args_list[0], ((d, ),))
            self.assertEqual(p_size.call_args_list[1], ((d, ), dict(exact=True)))


if __name__ == "__main__":
    unittest.main()



"""
=========
Visualize
=========

Widgets for data visualization.

"""

# Category description for the widget registry
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.visualize.visualize." + key)


NAME = "Visualize"

ID = "orange.widgets.visualize"

LABEL = __("package_label")

DESCRIPTION = __("package_desc")


BACKGROUND = "#FFB7B1"

ICON = "icons/Category-Visualize.svg"

PRIORITY = 3



from Orange.classification import SklTreeLearner
from Orange.classification import TreeLearner as ClassificationTreeLearner
from Orange.modelling import Fitter, SklFitter
from Orange.regression import TreeLearner as RegressionTreeLearner
from Orange.regression.tree import SklTreeRegressionLearner
from Orange.tree import TreeModel
from Orange.i18n_config import *
def __(key):
    return i18n.t("orange." + key)

__all__ = ['SklTreeLearner', 'TreeLearner']


class SklTreeLearner(SklFitter):
    name = __('name.tree')

    __fits__ = {'classification': SklTreeLearner,
                'regression': SklTreeRegressionLearner}


class TreeLearner(Fitter):
    name = __('name.tree')

    __fits__ = {'classification': ClassificationTreeLearner,
                'regression': RegressionTreeLearner}

    __returns__ = TreeModel



"""Learners"""
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.model.model." + key)


NAME = 'Model'
ID = "orange.widgets.model"
LABEL = __("package_label")

DESCRIPTION = __("package_desc")

BACKGROUND = '#FAC1D9'

ICON = 'icons/Category-Model.svg'

PRIORITY = 4



"""
============
Unsupervised
============

Unsupervised learning.

"""
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.unsupervised.unsupervised." + key)


# Category description for the widget registry

NAME = "Unsupervised"

LABEL = __("package_label")
ID = "orange.widgets.unsupervised"
DESCRIPTION = __("package_desc")

BACKGROUND = "#CAE1EF"

ICON = "icons/Category-Unsupervised.svg"

PRIORITY = 6



"""Naive Bayes Learner
"""

from Orange.data import Table
from Orange.classification.naive_bayes import NaiveBayesLearner
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.model.model.ownaivebayes." + key)

class OWNaiveBayes(OWBaseLearner):
    name = __("name")
    description = __("desc")
    icon = "icons/NaiveBayes.svg"
    replaces = [
        "Orange.widgets.classify.ownaivebayes.OWNaiveBayes",
    ]
    priority = 70
    keywords = "naive bayes"

    LEARNER = NaiveBayesLearner


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWNaiveBayes).run(Table("iris"))



# pylint: disable=missing-docstring,protected-access
import unittest
from unittest.mock import patch

import numpy as np

from AnyQt.QtCore import QEvent, QPointF, Qt
from AnyQt.QtGui import QMouseEvent
from AnyQt.QtTest import QSignalSpy

from Orange.data import Table, DiscreteVariable, Domain, ContinuousVariable, \
    StringVariable
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin
from Orange.widgets.visualize.owmosaic import OWMosaicDisplay
from Orange.widgets.tests.utils import simulate
from Orange.widgets.visualize.utils.vizrank import RunState


class TestOWMosaicDisplay(WidgetTest, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls)

        cls.signal_name = OWMosaicDisplay.Inputs.data
        cls.signal_data = cls.data

    def setUp(self):
        self.widget = self.create_widget(OWMosaicDisplay)

    def test_no_data(self):
        """Check that the widget doesn't crash on empty data"""
        self.send_signal(self.widget.Inputs.data, self.data[:0])

    def test_empty_column(self):
        """Check that the widget doesn't crash if the columns are empty"""
        self.send_signal(self.widget.Inputs.data, self.data[:, :0])

    def _select_data(self):
        self.widget.select_area(1, QMouseEvent(
            QEvent.MouseButtonPress, QPointF(), Qt.LeftButton,
            Qt.LeftButton, Qt.NoModifier))
        return [2, 3, 9, 23, 29, 30, 34, 35, 37, 42, 47, 49]

    def test_continuous_metas(self):
        """Check widget for dataset with continuous metas"""
        domain = Domain([ContinuousVariable("c1")],
                        metas=[ContinuousVariable("m")])
        data = Table(domain, np.arange(6).reshape(6, 1),
                     metas=np.arange(6).reshape(6, 1))
        self.send_signal(self.widget.Inputs.data, data)

    def test_string_meta(self):
        """Check widget for dataset with only one string meta"""
        domain = Domain([], metas=[StringVariable("m")])
        data = Table(domain, np.empty((6, 0)),
                     metas=np.array(["meta"] * 6).reshape(6, 1))
        self.send_signal(self.widget.Inputs.data, data)

    def test_missing_values(self):
        """Check widget for dataset with missing values"""
        data = Table(Domain([DiscreteVariable("c1", [])]),
                     np.array([np.nan] * 6)[:, None])
        self.send_signal(self.widget.Inputs.data, data)

        # missings in class variable
        attrs = [DiscreteVariable("c1", ["a", "b", "c"])]
        class_var = DiscreteVariable("cls", [])
        X = np.array([1, 2, 0, 1, 0, 2])[:, None]
        data = Table(Domain(attrs, class_var), X, np.array([np.nan] * 6))
        self.send_signal(self.widget.Inputs.data, data)

    def test_keyerror(self):
        """gh-2014
        Check if it works when a table has only one row or duplicates.
        Discretizer must have remove_const set to False.
        """
        data = Table("iris")
        data = data[0:1]
        self.send_signal(self.widget.Inputs.data, data)

    def test_combos_and_mosaic(self):
        """
        Text in combos is wiped away when there is no data and mosaic disappears as well.
        GH-2459
        GH-2462
        """
        def assertCount(cb_color, cb_attr, areas):
            self.assertEqual(len(self.widget.areas), areas)
            self.assertEqual(self.widget.cb_attr_color.count(), cb_color)
            for combo, cb_count in zip(self.widget.attr_combos, cb_attr):
                self.assertEqual(combo.count(), cb_count)

        data = Table("iris")
        assertCount(1, [0, 1, 1, 1], 0)
        self.send_signal(self.widget.Inputs.data, data)
        assertCount(6, [5, 6, 6, 6], 16)
        self.send_signal(self.widget.Inputs.data, None)
        assertCount(1, [0, 1, 1, 1], 0)

    @patch('Orange.widgets.visualize.owmosaic.CanvasRectangle')
    @patch('Orange.widgets.visualize.owmosaic.QGraphicsItemGroup.addToGroup')
    def test_different_number_of_attributes(self, _, canvas_rectangle):
        domain = Domain([DiscreteVariable(c, values="01") for c in "abcd"])
        data = Table.from_list(
            domain,
            [list("{:04b}".format(i)[-4:]) for i in range(16)])
        self.send_signal(self.widget.Inputs.data, data)
        widget = self.widget
        widget.variable2 = widget.variable3 = widget.variable4 = None
        for i, attr in enumerate(data.domain[:4], start=1):
            canvas_rectangle.reset_mock()
            setattr(self.widget, "variable" + str(i), attr)
            self.widget.update_graph()
            self.assertEqual(canvas_rectangle.call_count, 7 + 2 ** (i + 1))

    def test_change_domain(self):
        """Test for GH-3419 fix"""
        self.send_signal(self.widget.Inputs.data, self.data[:, :2])
        subset = self.data[:1, 2:3]
        self.send_signal(self.widget.Inputs.data, subset)
        output = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output.X, subset.X)

    def test_subset(self):
        """Test for GH-3416 fix"""
        self.send_signal(self.widget.Inputs.data, self.data)
        self.send_signal(self.widget.Inputs.data_subset, self.data[10:])
        self.send_signal(self.widget.Inputs.data, self.data[:1])
        output = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output.X, self.data[:1].X)

    @patch('Orange.widgets.visualize.owmosaic.MosaicVizRank.auto_select')
    def test_vizrank_receives_manual_change(self, auto_select):
        # Recreate the widget so the patch kicks in
        self.widget = self.create_widget(OWMosaicDisplay)
        data = Table("iris.tab")
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.variable1 = data.domain[0]
        self.widget.variable2 = data.domain[1]
        simulate.combobox_activate_index(self.widget.controls.variable2, 3)
        self.assertEqual(self.widget.variable2, data.domain[2])
        call_args = auto_select.call_args[0][0]
        self.assertEqual(len(call_args), 2)
        self.assertEqual(call_args[0].name, data.domain[0].name)
        self.assertEqual(call_args[1].name, data.domain[2].name)

    def test_selection_setting(self):
        widget = self.widget
        data = Table("iris.tab")
        self.send_signal(widget.Inputs.data, data)

        widget.select_area(
            1,
            QMouseEvent(QEvent.MouseButtonPress, QPointF(), Qt.LeftButton,
                        Qt.LeftButton, Qt.NoModifier))

        # Changing the data must reset the selection
        self.send_signal(widget.Inputs.data, Table("titanic"))
        self.assertFalse(bool(widget.selection))
        self.assertIsNone(self.get_output(widget.Outputs.selected_data))

        self.send_signal(widget.Inputs.data, data)
        self.assertFalse(bool(widget.selection))
        self.assertIsNone(self.get_output(widget.Outputs.selected_data))

        widget.select_area(
            1,
            QMouseEvent(QEvent.MouseButtonPress, QPointF(), Qt.LeftButton,
                        Qt.LeftButton, Qt.NoModifier))
        settings = self.widget.settingsHandler.pack_data(self.widget)

        # Setting data to None must reset the selection
        self.send_signal(widget.Inputs.data, None)
        self.assertFalse(bool(widget.selection))
        self.assertIsNone(self.get_output(widget.Outputs.selected_data))

        self.send_signal(widget.Inputs.data, data)
        self.assertFalse(bool(widget.selection))
        self.assertIsNone(self.get_output(widget.Outputs.selected_data))

        w = self.create_widget(OWMosaicDisplay, stored_settings=settings)
        self.assertFalse(bool(widget.selection))
        self.send_signal(w.Inputs.data, data, widget=w)
        self.assertEqual(w.selection, {1})
        self.assertIsNotNone(self.get_output(w.Outputs.selected_data, widget=w))


# Derive from WidgetTest to simplify creation of the Mosaic widget, although
# we are actually testing the MosaicVizRank dialog and not the widget

# These tests are rather crude: the main challenge of this widget is to handle
# user interactions and interrupts, e.g. changing the widget settings or
# getting new data while the computation is running.
class MosaicVizRankTests(WidgetTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.iris = Table("iris.tab")
        dom = Domain(cls.iris.domain.attributes, [])
        cls.iris_no_class = cls.iris.transform(dom)

    def setUp(self):
        self.widget = self.create_widget(OWMosaicDisplay)
        self.widget.vizrank_attr_range_index = 0

    def tearDown(self):
        self.widget.onDeleteWidget()
        super().tearDown()

    def test_count(self):
        """MosaicVizrank correctly computes the number of combinations"""

        data = self.iris
        attributes = [v for v in data.domain.attributes[1:]]
        metas = [data.domain.attributes[0]]
        domain = Domain(attributes, data.domain.class_var, metas)
        new_data = data.from_table(domain, data)

        for data in [self.iris, new_data]:
            self.send_signal(self.widget.Inputs.data, data)

            simulate.combobox_activate_index(self.widget.controls.variable_color, 0, 0)
            vizrank = self.widget.vizrank_dialog
            vizrank.attr_range_index = 1
            self.assertEqual(vizrank.state_count(), 10)  # 5x4 / 2
            vizrank.attr_range_index = 2
            self.assertEqual(vizrank.state_count(), 10)  # 5x4x3 / 2x3
            vizrank.attr_range_index = 3
            self.assertEqual(vizrank.state_count(), 5)  # 5x4x3x2 / 2x3x4
            vizrank.attr_range_index = 4
            self.assertEqual(vizrank.state_count(), 10)  # 5x4 / 2
            vizrank.attr_range_index = 5
            self.assertEqual(vizrank.state_count(), 20)  # above + 5x4x3 / 2x3
            vizrank.attr_range_index = 6
            self.assertEqual(vizrank.state_count(), 25)  # above + 5x4x3x2 / 2x3x4

            simulate.combobox_activate_index(self.widget.controls.variable_color, 2, 0)
            vizrank = self.widget.vizrank_dialog
            vizrank.attr_range_index = 0
            self.assertEqual(vizrank.state_count(), 4)  # 4
            vizrank.attr_range_index = 1
            self.assertEqual(vizrank.state_count(), 6)  # 4x3 / 2
            vizrank.attr_range_index = 2
            self.assertEqual(vizrank.state_count(), 4)  # 4x3x2 / 3x2
            vizrank.attr_range_index = 3
            self.assertEqual(vizrank.state_count(), 1)  # 4x3x2x1 / 2x3x4
            vizrank.attr_range_index = 4
            self.assertEqual(vizrank.state_count(), 10)  # 4 + 4x3 / 2
            vizrank.attr_range_index = 5
            self.assertEqual(vizrank.state_count(), 14)  # above + 4x3x2 / 3x2
            vizrank.attr_range_index = 6
            self.assertEqual(vizrank.state_count(), 15)  # above + 4x3x2x1 / 2x3x4

            self.send_signal(self.widget.Inputs.data, self.iris_no_class)
            simulate.combobox_activate_index(self.widget.controls.variable_color, 0, 0)
            vizrank = self.widget.vizrank_dialog
            vizrank.attr_range_index = 1
            self.assertEqual(vizrank.state_count(), 6)  # 4x3 / 2
            vizrank.attr_range_index = 2
            self.assertEqual(vizrank.state_count(), 4)  # 4x3x2 / 3x2
            vizrank.attr_range_index = 3
            self.assertEqual(vizrank.state_count(), 1)  # 4x3x2x1 / 2x3x4
            vizrank.attr_range_index = 4
            self.assertEqual(vizrank.state_count(), 6)  # 4x3 / 2
            vizrank.attr_range_index = 5
            self.assertEqual(vizrank.state_count(), 10)  # above + 4x3x2 / 3x2
            vizrank.attr_range_index = 6
            self.assertEqual(vizrank.state_count(), 11)  # above + 4x3x2x1 / 2x3x4

    def test_iteration(self):
        """MosaicVizrank correctly iterates through states"""
        widget = self.widget
        widget.vizrank_attr_range_index = 1
        self.send_signal(self.widget.Inputs.data, self.iris)
        vizrank = self.widget.vizrank_dialog

        self.assertEqual(set(vizrank.state_generator()),
                         {(0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3)})

        vizrank.attr_range_index = 6
        self.assertEqual(set(vizrank.state_generator()),
                         {(0, ), (1, ), (2, ), (3, ),
                          (0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3),
                          (0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3),
                          (0, 1, 2, 3)})

        vizrank.attr_range_index = 4
        self.assertEqual(set(vizrank.state_generator()),
                         {(0, ), (1, ), (2, ), (3, ),
                          (0, 1), (0, 2), (1, 2), (0, 3), (1, 3), (2, 3)})

        widget.vizrank_attr_range_index = 6
        widget.variable_color = None
        self.widget.cb_attr_color.setCurrentIndex(0)
        simulate.combobox_activate_index(self.widget.controls.variable_color, 0, 0)
        vizrank = self.widget.vizrank_dialog
        self.assertEqual(set(vizrank.state_generator()),
                         {(0, 1), (0, 2), (0, 3), (0, 4),
                          (1, 2), (1, 3), (1, 4),
                          (2, 3), (2, 4),
                          (3, 4),
                          (0, 1, 2), (0, 1, 3), (0, 1, 4), (0, 2, 3), (0, 2, 4), (0, 3, 4),
                          (1, 2, 3), (1, 2, 4), (1, 3, 4),
                          (2, 3, 4),
                          (0, 1, 2, 3), (0, 1, 3, 4), (0, 1, 2, 4), (0, 2, 3, 4), (1, 2, 3, 4)})

        vizrank.attr_range_index = 4
        self.assertEqual(set(vizrank.state_generator()),
                         {(0, 1), (0, 2), (0, 3), (0, 4),
                          (1, 2), (1, 3), (1, 4),
                          (2, 3), (2, 4),
                          (3, 4)})

    def test_row_for_state(self):
        """MosaicVizrank returns table row corresponding to the state"""
        self.send_signal(self.widget.Inputs.data, self.iris)
        vizrank = self.widget.vizrank_dialog
        vizrank._attr_order = [DiscreteVariable(n) for n in "abcd"]
        items = vizrank.row_for_state(0, [1, 3, 0])
        self.assertEqual(len(items), 1)
        item = items[0]
        self.assertEqual(item.text(), "a, b, d")
        self.assertEqual(
            item.data(vizrank._AttrRole),
            tuple(vizrank.attr_order[i] for i in [0, 1, 3]))

    def test_does_not_crash_cont_class(self):
        """MosaicVizrank computes rankings without crashing"""
        data = Table("housing.tab")
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.vizrank_button().click()

    def test_max_attr_combo_1_disabling(self):
        widget = self.widget
        enabled = Qt.ItemIsSelectable | Qt.ItemIsEnabled

        data = Table("iris.tab")
        self.send_signal(self.widget.Inputs.data, data)
        vizrank = widget.vizrank_dialog
        combo = vizrank.attrs_combo
        model = combo.model()
        self.assertEqual(model.item(0).flags() & enabled, enabled)

        vizrank.attr_range_index = 0
        simulate.combobox_activate_index(self.widget.controls.variable_color, 0)
        vizrank = widget.vizrank_dialog
        combo = vizrank.attrs_combo
        model = combo.model()
        self.assertEqual(vizrank.attr_range_index, 1)
        self.assertEqual(model.item(0).flags() & enabled, Qt.NoItemFlags)

        simulate.combobox_activate_index(self.widget.controls.variable_color, 1)
        vizrank = widget.vizrank_dialog
        combo = vizrank.attrs_combo
        model = combo.model()
        self.assertEqual(model.item(0).flags() & enabled, enabled)

    def test_disable_combo_3_4(self):
        def assert_enabled(*args, sel=1):
            enabled = Qt.ItemIsSelectable | Qt.ItemIsEnabled
            vizrank = widget.vizrank_dialog
            combo = vizrank.attrs_combo
            model = combo.model()
            for opt in (2, 3):
                self.assertEqual(model.item(opt).flags() & enabled,
                                 enabled if opt + 1 in args else Qt.NoItemFlags)
            self.assertEqual(combo.currentIndex(), sel)

        widget = self.widget

        data = Table("iris.tab")
        widget.vizrank_attr_range_index = 3
        self.send_signal(self.widget.Inputs.data, data)
        assert_enabled(3, 4, sel=3)

        self.send_signal(self.widget.Inputs.data, data[:, :3])
        assert_enabled(3, sel=2)

        self.send_signal(self.widget.Inputs.data, data[:, :2])
        assert_enabled(sel=1)

    def test_attr_range(self):
        data = Table("iris.tab")
        domain = data.domain

        self.send_signal(self.widget.Inputs.data, data)
        vizrank = self.widget.vizrank_dialog
        for vizrank.attr_range_index, rge in (
                (0, (1, 2)), (1, (2, 3)), (2, (3, 4)), (3, (4, 5)),
                (4, (1, 3)), (5, (1, 4)), (6, (1, 5))):
            self.assertEqual(vizrank.attr_range(), rge,
                             f"failed at attr_range_index={vizrank.attr_range_index}")

        reduced = data.transform(Domain(domain.attributes[:2], domain.class_var))
        self.send_signal(self.widget.Inputs.data, reduced)
        vizrank = self.widget.vizrank_dialog
        for vizrank.attr_range_index, rge in (
                (0, (1, 2)), (1, (2, 3)), (2, (3, 3)), (3, (4, 3)),
                (4, (1, 3)), (5, (1, 3)), (6, (1, 3))):
            self.assertEqual(vizrank.attr_range(), rge,
                             f"failed at attr_range_index={vizrank.attr_range_index}")
            self.assertIs(vizrank.state_count() == 0, rge[0] >= rge[1])

        simulate.combobox_activate_index(self.widget.controls.variable_color, 0)
        vizrank = self.widget.vizrank_dialog
        for vizrank.attr_range_index, rge in (
                (0, (2, 3)), (1, (2, 3)), (2, (3, 4)), (3, (4, 4)),
                (4, (2, 3)), (5, (2, 4)), (6, (2, 4))):
            self.assertEqual(vizrank.attr_range(), rge,
                             f"failed at attr_range_index={vizrank.attr_range_index}")
            self.assertIs(vizrank.state_count() == 0, rge[0] >= rge[1])

    def test_nan_column(self):
        """
        A column with only NaN-s used to throw an error
        (ZeroDivisionError) when loaded into widget.
        GH-2046
        """
        table = Table(
            Domain(
                [ContinuousVariable("a"), ContinuousVariable("b"), ContinuousVariable("c")]),
            np.array([
                [0, np.NaN, 0],
                [0, np.NaN, 0],
                [0, np.NaN, 0]
            ])
        )
        self.send_signal(self.widget.Inputs.data, table)

    def test_color_combo(self):
        """
        Color combo enables to select class values. Checks if class values
        are selected correctly.
        """
        table = Table("titanic")
        self.send_signal(self.widget.Inputs.data, table)
        color_vars = ["(Pearson residuals)"] + [str(x) for x in table.domain.variables]
        for cv in color_vars:
            idx = self.widget.cb_attr_color.findText(cv)
            self.widget.cb_attr_color.setCurrentIndex(idx)
            color = self.widget.cb_attr_color.currentText()
            simulate.combobox_activate_index(self.widget.controls.variable_color, idx, 0)
            discrete_data = self.widget.discrete_data

            if color == "(Pearson residuals)":
                self.assertIsNone(discrete_data.domain.class_var)
            else:
                self.assertEqual(color, str(discrete_data.domain.class_var))

            output = self.get_output(self.widget.Outputs.annotated_data)
            self.assertEqual(output.domain.class_var, table.domain.class_var)

    def test_scores(self):
        SCORES = {('status', ): 4.35e-40,
                  ('sex', ): 6.18e-100,
                  ('age', ): 2.82e-05,
                  ('sex', 'status'): 1.06e-123,
                  ('age', 'status'): 4.15e-47,
                  ('age', 'sex'): 5.49e-102,
                  ('age', 'sex', 'status'): 5.3e-128}
        table = Table("titanic")
        self.send_signal(self.widget.Inputs.data, table)
        vizrank = self.widget.vizrank_dialog
        vizrank.prepare_run()
        vizrank.attr_range_index = 3
        for state in vizrank.state_generator():
            attrlist = tuple(sorted(vizrank.attr_order[i].name for i in state))
            sc = vizrank.compute_score(state)
            self.assertTrue(np.allclose(sc, SCORES[attrlist], rtol=0.003, atol=0))

    def test_subset_data(self):
        """
        Mosaic should not crash on subset data and should properly interpret it.
        GH-2515
        GH-2528
        """
        table_titanic = Table("titanic")
        self.send_signal(self.widget.Inputs.data, table_titanic)
        self.send_signal(self.widget.Inputs.data_subset, table_titanic[::20])
        table_housing = Table("housing")
        self.send_signal(self.widget.Inputs.data, table_housing)
        self.send_signal(self.widget.Inputs.data_subset, table_housing[::20])

    def test_incompatible_subset(self):
        """
        Show warning when subset data is not compatible with data.
        GH-2528
        """
        table_titanic = Table("titanic")
        self.assertFalse(self.widget.Warning.incompatible_subset.is_shown())
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.send_signal(self.widget.Inputs.data_subset, table_titanic)
        self.assertTrue(self.widget.Warning.incompatible_subset.is_shown())
        self.send_signal(self.widget.Inputs.data_subset, self.iris)
        self.assertFalse(self.widget.Warning.incompatible_subset.is_shown())

    def test_autoselect(self):
        data = Table("iris.tab")
        self.widget.vizrank_attr_range_index = 2
        self.send_signal(self.widget.Inputs.data, data)
        spy = QSignalSpy(self.widget.vizrankRunStateChanged)
        self.widget.vizrank_button().click()
        # This takes 0.5 s on my M1 Mac (2022), but let us tolerate 20x longer
        # on CI
        while spy.wait(timeout=10000) and spy[-1][0] != RunState.Done:
            pass

        vizrank = self.widget.vizrank_dialog
        model = vizrank.rank_model
        attrs = model.data(model.index(3, 0), vizrank._AttrRole)
        vizrank.auto_select(attrs)
        selection = vizrank.rank_table.selectedIndexes()
        self.assertEqual(len(selection), 1)
        self.assertEqual(selection[0].row(), 3)

        vizrank.auto_select(attrs[::-1])
        selection = vizrank.rank_table.selectedIndexes()
        self.assertEqual(len(selection), 0)


if __name__ == "__main__":
    unittest.main()



from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import (
    QFormLayout, QCheckBox, QLineEdit, QWidget, QVBoxLayout, QLabel
)
from orangecanvas.application.settings import UserSettingsDialog, FormLayout
from orangecanvas.document.interactions import PluginDropHandler
from orangecanvas.document.usagestatistics import UsageStatistics
from orangecanvas.utils.overlay import NotificationOverlay

from orangewidget.workflow.mainwindow import OWCanvasMainWindow

from Orange.i18n_config import *
def __(key):
    return i18n.t("orange." + key)


class OUserSettingsDialog(UserSettingsDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        w = self.widget(0)  # 'General' tab
        layout = w.layout()
        assert isinstance(layout, QFormLayout)
        cb = QCheckBox(self.tr(__("checkbox.auto_update")))
        cb.setAttribute(Qt.WA_LayoutUsesWidgetRect)

        layout.addRow(__("btn.update"), cb)
        self.bind(cb, "checked", "startup/check-updates")

        # Reporting Tab
        tab = QWidget()
        self.addTab(tab, self.tr(__("btn.report")),
                    toolTip=__("btn.report_tip"))

        form = FormLayout()
        line_edit_mid = QLineEdit()
        self.bind(line_edit_mid, "text", "reporting/machine-id")
        form.addRow(__("row.machine_id"), line_edit_mid)

        box = QWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        cb1 = QCheckBox(
            self.tr(__("checkbox.share")),
            toolTip=self.tr(__("checkbox.share_tip"))
        )
        self.bind(cb1, "checked", "reporting/send-statistics")
        cb1.clicked.connect(UsageStatistics.set_enabled)
        layout.addWidget(cb1)
        box.setLayout(layout)
        form.addRow(self.tr(__("row.anonymous_statistic")), box)
        # label = QLabel(__("label.more_info"))
        # label.setOpenExternalLinks(True)
        # form.addRow(self.tr(""), label)

        tab.setLayout(form)

        # Notifications Tab
        tab = QWidget()
        self.addTab(tab, self.tr(__("btn.notification")),
                    toolTip=__("btn.notification_tip"))

        form = FormLayout()

        box = QWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        cb = QCheckBox(
            self.tr(__("checkbox.enable_notification")), self,
            toolTip=__("checkbox.enable_notification_tip")
        )
        self.bind(cb, "checked", "notifications/check-notifications")

        layout.addWidget(cb)
        box.setLayout(layout)
        form.addRow(self.tr(__("row.startup")), box)

        notifs = QWidget(self, objectName="notifications-group")
        notifs.setLayout(QVBoxLayout())
        notifs.layout().setContentsMargins(0, 0, 0, 0)

        cb1 = QCheckBox(self.tr(__("checkbox.announcement")), self,
                        toolTip=__("checkbox.announcement_tip"))

        cb2 = QCheckBox(self.tr(__("checkbox.blog_post")), self,
                        toolTip=__("checkbox.blog_post_tip"))
        cb3 = QCheckBox(self.tr(__("checkbox.new_feature")), self,
                        toolTip=__("checkbox.new_feature_tip"))

        self.bind(cb1, "checked", "notifications/announcements")
        self.bind(cb2, "checked", "notifications/blog")
        self.bind(cb3, "checked", "notifications/new-features")

        notifs.layout().addWidget(cb1)
        notifs.layout().addWidget(cb2)
        notifs.layout().addWidget(cb3)

        form.addRow(self.tr(__("row.notification_about")), notifs)
        tab.setLayout(form)


class MainWindow(OWCanvasMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notification_overlay = NotificationOverlay(self.scheme_widget)
        self.notification_server = None
        self.scheme_widget.setDropHandlers([
            PluginDropHandler("orange.canvas.drophandler")
        ])

    def open_canvas_settings(self):
        # type: () -> None
        """Reimplemented."""
        dlg = OUserSettingsDialog(self, windowTitle=self.tr(__("menu.perferences")))
        dlg.show()
        status = dlg.exec()
        if status == 0:
            self.user_preferences_changed_notify_all()

    def set_notification_server(self, notif_server):
        self.notification_server = notif_server

        # populate notification overlay with current notifications
        for notif in self.notification_server.getNotificationQueue():
            self.notification_overlay.addNotification(notif)

        notif_server.newNotification.connect(self.notification_overlay.addNotification)
        notif_server.nextNotification.connect(self.notification_overlay.nextWidget)

    def create_new_window(self):  # type: () -> CanvasMainWindow
        window = super().create_new_window()
        window.set_notification_server(self.notification_server)
        return window



from Orange.classification import CalibratedLearner, ThresholdLearner, \
    NaiveBayesLearner
from Orange.data import Table
from Orange.i18n_config import *
from Orange.modelling import Learner
from Orange.widgets import gui
from Orange.widgets.widget import Input
from Orange.widgets.settings import Setting
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.widgetpreview import WidgetPreview


def __(key):
    return i18n.t("widget.model.model.owcalibratedlearner." + key)


class OWCalibratedLearner(OWBaseLearner):
    name = __("name")
    description = __("desc")
    icon = "icons/CalibratedLearner.svg"
    priority = 20
    keywords = "calibrated learner, calibration, threshold"

    LEARNER = CalibratedLearner

    SigmoidCalibration, IsotonicCalibration, NoCalibration = range(3)
    CalibrationOptions = (i18n.t("common.algorithm.sigmoid_calibration"),
                          i18n.t("common.algorithm.isotonic_calibration"),
                          __("btn.no_calibration"))
    CalibrationShort = ("Sigmoid", "Isotonic", "")
    CalibrationMap = {
        SigmoidCalibration: CalibratedLearner.Sigmoid,
        IsotonicCalibration: CalibratedLearner.Isotonic}

    OptimizeCA, OptimizeF1, NoThresholdOptimization = range(3)
    ThresholdOptions = (__("btn.optimize_class_accuracy"),
                        __("btn.optimize_f1_score"),
                        __("btn.no_threshold_optimization"))
    ThresholdShort = ("CA", "F1", "")
    ThresholdMap = {
        OptimizeCA: ThresholdLearner.OptimizeCA,
        OptimizeF1: ThresholdLearner.OptimizeF1}

    learner_name = Setting("", schema_only=True)
    calibration = Setting(SigmoidCalibration)
    threshold = Setting(OptimizeCA)

    class Inputs(OWBaseLearner.Inputs):
        base_learner = Input("Base Learner", Learner, label=i18n.t("widget.model.model.common.base_learner"))

    def __init__(self):
        super().__init__()
        self.base_learner = None

    def add_main_layout(self):
        gui.radioButtons(
            self.controlArea, self, "calibration", self.CalibrationOptions,
            box=__("box_probability_calibration"),
            callback=self.calibration_options_changed)
        gui.radioButtons(
            self.controlArea, self, "threshold", self.ThresholdOptions,
            box=__("box_decision_threshold_optimization"),
            callback=self.calibration_options_changed)

    @Inputs.base_learner
    def set_learner(self, learner):
        self.base_learner = learner
        self._set_default_name()
        self.learner = self.model = None

    def _set_default_name(self):

        if self.base_learner is None:
            self.set_default_learner_name("")
        else:
            name = " + ".join(part for part in (
                self.base_learner.name.title(),
                self.CalibrationShort[self.calibration],
                self.ThresholdShort[self.threshold]) if part)
            self.set_default_learner_name(name)

    def calibration_options_changed(self):
        self._set_default_name()
        self.apply()

    def create_learner(self):
        class IdentityWrapper(Learner):
            def fit_storage(self, data):
                return self.base_learner.fit_storage(data)

        if self.base_learner is None:
            return None
        learner = self.base_learner
        if self.calibration != self.NoCalibration:
            learner = CalibratedLearner(learner,
                                        self.CalibrationMap[self.calibration])
        if self.threshold != self.NoThresholdOptimization:
            learner = ThresholdLearner(learner,
                                       self.ThresholdMap[self.threshold])
        if self.preprocessors:
            if learner is self.base_learner:
                learner = IdentityWrapper()
            learner.preprocessors = (self.preprocessors,)
        return learner

    def get_learner_parameters(self):
        return ((__("report.calibrate_probability"),
                 self.CalibrationOptions[self.calibration]),
                (__("report.threshold_optimization"),
                 self.ThresholdOptions[self.threshold]))


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWCalibratedLearner).run(
        Table("heart_disease"),
        set_learner=NaiveBayesLearner())



from AnyQt.QtCore import QUrl, QMimeData

from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import dragDrop
from Orange.widgets.utils.filedialogs import OWUrlDropBase


class TestOWUrlDropBase(WidgetTest):
    def test_drop(self):
        class TestW(OWUrlDropBase):
            path = None

            def canDropUrl(self, url: QUrl) -> bool:
                return url.toLocalFile().endswith(".foo")

            def handleDroppedUrl(self, url: QUrl) -> None:
                self.path = url.toLocalFile()

        w = self.create_widget(TestW)
        url = QUrl("file:///bar.foo")
        mime = QMimeData()
        mime.setUrls([url])
        self.assertTrue(dragDrop(w, mime))
        self.assertEqual(w.path, url.toLocalFile())
        url = QUrl("file:///bar.baz")
        mime.setUrls([url])
        self.assertFalse(dragDrop(w, mime))
        self.assertNotEqual(w.path, url.toLocalFile())



"""Tree learner widget"""

from collections import OrderedDict

from AnyQt.QtCore import Qt

from Orange.data import Table
from Orange.modelling.tree import TreeLearner
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.utils.localization import pl
from Orange.widgets.utils.owlearnerwidget import OWBaseLearner
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.i18n_config import *



def __(key):
    return i18n.t("widget.model.model.owtree." + key)

class OWTreeLearner(OWBaseLearner):
    """Tree algorithm with forward pruning."""
    name = __("name")
    description = __("desc")
    icon = "icons/Tree.svg"
    replaces = [
        "Orange.widgets.classify.owclassificationtree.OWClassificationTree",
        "Orange.widgets.regression.owregressiontree.OWRegressionTree",
        "Orange.widgets.classify.owclassificationtree.OWTreeLearner",
        "Orange.widgets.regression.owregressiontree.OWTreeLearner",
    ]
    priority = 30
    keywords = "tree, classification tree"

    LEARNER = TreeLearner

    binary_trees = Setting(True)
    limit_min_leaf = Setting(True)
    min_leaf = Setting(2)
    limit_min_internal = Setting(True)
    min_internal = Setting(5)
    limit_depth = Setting(True)
    max_depth = Setting(100)

    # Classification only settings
    limit_majority = Setting(True)
    sufficient_majority = Setting(95)

    spin_boxes = (
        (__("checkbox.leave_min_num"),
         "limit_min_leaf", "min_leaf", 1, 1000),
        (__("checkbox.not_split_smaller_subset"),
         "limit_min_internal", "min_internal", 1, 1000),
        (__("checkbox.limit_max_tree_depth"),
         "limit_depth", "max_depth", 1, 1000))

    classification_spin_boxes = (
        (__("checkbox.majority_reaches_stop"),
         "limit_majority", "sufficient_majority", 51, 100),)

    def add_main_layout(self):
        box = gui.widgetBox(self.controlArea, i18n.t("common.general.parameter"))
        # the checkbox is put into vBox for alignemnt with other checkboxes
        gui.checkBox(box, self, "binary_trees", __("checkbox.induce_binary_tree"),
                     callback=self.settings_changed,
                     attribute=Qt.WA_LayoutUsesWidgetRect)
        for label, check, setting, fromv, tov in self.spin_boxes:
            gui.spin(box, self, setting, fromv, tov, label=label,
                     checked=check, alignment=Qt.AlignRight,
                     callback=self.settings_changed,
                     checkCallback=self.settings_changed, controlWidth=80)

    def add_classification_layout(self, box):
        for label, check, setting, minv, maxv in self.classification_spin_boxes:
            gui.spin(box, self, setting, minv, maxv,
                     label=label, checked=check, alignment=Qt.AlignRight,
                     callback=self.settings_changed, controlWidth=80,
                     checkCallback=self.settings_changed)

    def learner_kwargs(self):
        # Pylint doesn't get our Settings
        # pylint: disable=invalid-sequence-index
        return dict(
            max_depth=(None, self.max_depth)[self.limit_depth],
            min_samples_split=(2, self.min_internal)[self.limit_min_internal],
            min_samples_leaf=(1, self.min_leaf)[self.limit_min_leaf],
            binarize=self.binary_trees,
            preprocessors=self.preprocessors,
            sufficient_majority=(1, self.sufficient_majority / 100)[
                self.limit_majority])

    def create_learner(self):
        # pylint: disable=not-callable
        return self.LEARNER(**self.learner_kwargs())

    def get_learner_parameters(self):
        from Orange.widgets.report import plural_w
        items = OrderedDict()
        items[__("report.prune")] = ", ".join(s for s, c in (
            (plural_w(__("report.number_instance_leaves"),
                      self.min_leaf), self.limit_min_leaf),
            (plural_w(__("report.number_instance_internal_nodes"),
                      self.min_internal), self.limit_min_internal),
            (__("report.maximum_depth").format(self.max_depth), self.limit_depth)
        ) if c) or i18n.t("common.software.none")
        if self.limit_majority:
            items[__("report.split")] = __("report.classification_stop_split") %(self.sufficient_majority, "%")
        items[__("report.binary_trees")] = (__("report.cancel"), __("report.ok"))[self.binary_trees]
        return items


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWTreeLearner).run(Table("iris"))



# pylint: disable=missing-docstring
import unittest

import numpy as np
from sklearn.cross_decomposition import PLSRegression

from Orange.data import Table, Domain, ContinuousVariable
from Orange.regression import PLSRegressionLearner
from Orange.regression.pls import _PLSCommonTransform


def table(rows, attr, variables):
    attr_vars = [ContinuousVariable(name=f"Feature {i}") for i in
                 range(attr)]
    class_vars = [ContinuousVariable(name=f"Class {i}") for i in
                  range(variables)]
    domain = Domain(attr_vars, class_vars, [])
    X = np.random.RandomState(0).random((rows, attr))
    Y = np.random.RandomState(1).random((rows, variables))
    return Table.from_numpy(domain, X=X, Y=Y)


class TestPLSRegressionLearner(unittest.TestCase):
    def test_allow_y_dim(self):
        """ The current PLS version allows only a single Y dimension. """
        learner = PLSRegressionLearner(n_components=2)
        d = table(10, 5, 0)
        with self.assertRaises(ValueError):
            learner(d)
        for n_class_vars in [1, 2, 3]:
            d = table(10, 5, n_class_vars)
            learner(d)  # no exception

    def test_compare_to_sklearn(self):
        d = table(10, 5, 1)
        orange_model = PLSRegressionLearner()(d)
        scikit_model = PLSRegression().fit(d.X, d.Y)
        np.testing.assert_almost_equal(scikit_model.predict(d.X).ravel(),
                                       orange_model(d))
        np.testing.assert_almost_equal(scikit_model.coef_,
                                       orange_model.coefficients)

    def test_compare_to_sklearn_multid(self):
        d = table(10, 5, 3)
        orange_model = PLSRegressionLearner()(d)
        scikit_model = PLSRegression().fit(d.X, d.Y)
        np.testing.assert_almost_equal(scikit_model.predict(d.X),
                                       orange_model(d))
        np.testing.assert_almost_equal(scikit_model.coef_,
                                       orange_model.coefficients)

    def test_too_many_components(self):
        # do not change n_components
        d = table(5, 5, 1)
        model = PLSRegressionLearner(n_components=4)(d)
        self.assertEqual(model.skl_model.n_components, 4)
        # need to use fewer components; column limited
        d = table(6, 5, 1)
        model = PLSRegressionLearner(n_components=6)(d)
        self.assertEqual(model.skl_model.n_components, 4)
        # need to use fewer components; row limited
        d = table(5, 6, 1)
        model = PLSRegressionLearner(n_components=6)(d)
        self.assertEqual(model.skl_model.n_components, 4)

    def test_scores(self):
        for d in [table(10, 5, 1), table(10, 5, 3)]:
            orange_model = PLSRegressionLearner()(d)
            scikit_model = PLSRegression().fit(d.X, d.Y)
            scores = orange_model.project(d)
            sx, sy = scikit_model.transform(d.X, d.Y)
            np.testing.assert_almost_equal(sx, scores.X)
            np.testing.assert_almost_equal(sy, scores.metas)

    def test_components(self):
        def t2d(m):
            return m.reshape(-1, 1) if len(m.shape) == 1 else m

        for d in [table(10, 5, 1), table(10, 5, 3)]:
            orange_model = PLSRegressionLearner()(d)
            scikit_model = PLSRegression().fit(d.X, d.Y)
            components = orange_model.components()
            np.testing.assert_almost_equal(scikit_model.x_loadings_,
                                           components.X.T)
            np.testing.assert_almost_equal(scikit_model.y_loadings_,
                                           t2d(components.Y).T)

    def test_coefficients(self):
        for d in [table(10, 5, 1), table(10, 5, 3)]:
            orange_model = PLSRegressionLearner()(d)
            scikit_model = PLSRegression().fit(d.X, d.Y)
            coef_table = orange_model.coefficients_table()
            np.testing.assert_almost_equal(scikit_model.coef_.T,
                                           coef_table.X)

    def test_residuals_normal_probability(self):
        for d in [table(10, 5, 1), table(10, 5, 3)]:
            orange_model = PLSRegressionLearner()(d)
            res_table = orange_model.residuals_normal_probability(d)
            n_target = len(d.domain.class_vars)
            self.assertEqual(res_table.X.shape, (len(d), 2 * n_target))

    def test_dmodx(self):
        for d in (table(10, 5, 1), table(10, 5, 3)):
            orange_model = PLSRegressionLearner()(d)
            dist_table = orange_model.dmodx(d)
            self.assertEqual(dist_table.X.shape, (len(d), 1))

    def test_eq_hash(self):
        data = Table("housing")
        pls1 = PLSRegressionLearner()(data)
        pls2 = PLSRegressionLearner()(data)

        proj1 = pls1.project(data)
        proj2 = pls2.project(data)

        np.testing.assert_equal(proj1.X, proj2.X)
        np.testing.assert_equal(proj1.metas, proj2.metas)

        # even though results are the same, these transformations
        # are different because the PLS object is
        self.assertNotEqual(proj1, proj2)
        self.assertNotEqual(proj1.domain, proj2.domain)
        self.assertNotEqual(hash(proj1), hash(proj2))
        self.assertNotEqual(hash(proj1.domain), hash(proj2.domain))

        proj2.domain[0].compute_value.compute_shared.pls_model = \
            proj1.domain[0].compute_value.compute_shared.pls_model
        # reset hash caches because object were hacked
        # pylint: disable=protected-access
        proj1.domain._hash = None
        proj2.domain._hash = None

        self.assertEqual(proj1.domain, proj2.domain)
        self.assertEqual(hash(proj1.domain), hash(proj2.domain))


class TestPLSCommonTransform(unittest.TestCase):
    def test_eq(self):
        m = PLSRegressionLearner()(table(10, 5, 1))
        transformer = _PLSCommonTransform(m)
        self.assertEqual(transformer, transformer)
        self.assertEqual(transformer, _PLSCommonTransform(m))

        m = PLSRegressionLearner()(table(10, 5, 2))
        self.assertNotEqual(transformer, _PLSCommonTransform(m))

    def test_hash(self):
        m = PLSRegressionLearner()(table(10, 5, 1))
        transformer = _PLSCommonTransform(m)
        self.assertEqual(hash(transformer), hash(transformer))
        self.assertEqual(hash(transformer), hash(_PLSCommonTransform(m)))

        m = PLSRegressionLearner()(table(10, 5, 2))
        self.assertNotEqual(hash(transformer), hash(_PLSCommonTransform(m)))

    def test_missing_target(self):
        data = table(10, 5, 1)
        with data.unlocked(data.Y):
            data.Y[::3] = np.nan
        pls = PLSRegressionLearner()(data)
        proj = pls.project(data)
        self.assertFalse(np.isnan(proj.X).any())
        self.assertFalse(np.isnan(proj.metas[1::3]).any())
        self.assertFalse(np.isnan(proj.metas[2::3]).any())
        self.assertTrue(np.isnan(proj.metas[::3]).all())

    def test_missing_target_multitarget(self):
        data = table(10, 5, 3)
        with data.unlocked(data.Y):
            data.Y[0] = np.nan
            data.Y[1, 1] = np.nan

        pls = PLSRegressionLearner()(data)
        proj = pls.project(data)
        self.assertFalse(np.isnan(proj.X).any())
        self.assertFalse(np.isnan(proj.metas[2:]).any())
        self.assertTrue(np.isnan(proj.metas[:2]).all())

    def test_apply_domain_classless_data(self):
        data = Table("housing")
        pls = PLSRegressionLearner()(data)
        classless_data = data.transform(Domain(data.domain.attributes))[:5]

        proj = pls.project(classless_data)
        self.assertFalse(np.isnan(proj.X).any())
        self.assertTrue(np.isnan(proj.metas).all())


if __name__ == "__main__":
    unittest.main()



"""
Linear Projection widget
------------------------
"""

from itertools import islice, permutations, chain, combinations
from math import factorial, comb

import numpy as np

from sklearn.neighbors import NearestNeighbors
from sklearn.metrics import r2_score

from AnyQt.QtGui import QPalette
from AnyQt.QtCore import QRectF, QLineF

import pyqtgraph as pg

from Orange.data import Table, Domain, IsDefined
from Orange.preprocess import Normalize
from Orange.preprocess.score import ReliefF, RReliefF
from Orange.projection import PCA, LDA, LinearProjector
from Orange.util import Enum
from Orange.widgets import gui, report
from Orange.widgets.settings import Setting, ContextSetting, SettingProvider
from Orange.widgets.utils.localization import pl
from Orange.widgets.utils.plot import variables_selection
from Orange.widgets.utils.plot.owplotgui import VariableSelectionModel
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.visualize.utils import vizrank
from Orange.widgets.visualize.utils.vizrank import VizRankDialogNAttrs, \
    VizRankMixin
from Orange.widgets.visualize.utils.component import OWGraphWithAnchors
from Orange.widgets.visualize.utils.plotutils import AnchorItem
from Orange.widgets.visualize.utils.widget import OWAnchorProjectionWidget
from Orange.widgets.widget import Msg
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.visualize.visualize.owlinearprojection." + key)


MAX_LABEL_LEN = 20


class LinearProjectionVizRank(VizRankDialogNAttrs):
    minK = 10
    show_bars = False

    def score_attributes(self):
        def normalized(a):
            span = np.max(a, axis=0) - np.min(a, axis=0)
            span[span == 0] = 1
            return (a - np.mean(a, axis=0)) / span

        domain = Domain(
            attributes=[v for v in self.attrs if v is not self.attr_color],
            class_vars=self.attr_color
        )
        data = self.data.transform(domain).copy()
        with data.unlocked():
            data.X = normalized(data.X)
        relief = ReliefF if self.attr_color.is_discrete else RReliefF
        weights = relief(n_iterations=100, k_nearest=self.minK)(data)
        results = sorted(zip(weights, domain.attributes),
                         key=lambda x: (-x[0], x[1].name))
        return [attr for _, attr in results]

    def state_count(self):
        n_all_attrs = self.max_attrs()
        if not n_all_attrs:
            return 0
        return comb(n_all_attrs, self.n_attrs) * factorial(self.n_attrs - 1) // 2

    def state_generator(self):
        return (
            (c[0], *p)
            for c in combinations(list(range(len(self.attr_order))), self.n_attrs)
            for p in islice(permutations(c[1:]), factorial(len(c) - 1) // 2)
        )

    def compute_score(self, state):
        domain = Domain([self.attr_order[i] for i in state], [self.attr_color])
        reduced = IsDefined()(self.data.transform(domain))
        if len(reduced) < self.minK:  # cancel early if not enough data
            return None
        projection = self.parent().projector(reduced)
        ec = projection(reduced).X
        if ec.shape[0] < self.minK:  # projection preprocessors can remove data(?)
            return None
        n_neighbors = min(self.minK, len(ec) - 1)
        knn = NearestNeighbors(n_neighbors=n_neighbors).fit(ec)
        ind = knn.kneighbors(return_distance=False)
        y = reduced.get_column(self.attr_color)
        if self.attr_color.is_discrete:
            score = -np.sum(y[ind] == y.reshape(-1, 1)) / n_neighbors
        else:
            score = -r2_score(y, np.mean(y[ind], axis=1))
        # treat missing data as misclassified
        return score * len(reduced) / len(self.data)


class OWLinProjGraph(OWGraphWithAnchors):
    hide_radius = Setting(0)

    @property
    def always_show_axes(self):
        return self.master.placement == Placement.Circular

    @property
    def scaled_radius(self):
        return self.hide_radius / 100 + 1e-5

    def update_radius(self):
        self.update_circle()
        self.update_anchors()

    def set_view_box_range(self):
        def min_max(a, b):
            return (min(np.amin(a), np.amin(b), -1.05),
                    max(np.amax(a), np.amax(b), 1.05))

        points, _ = self.master.get_anchors()
        coords = self.master.get_coordinates_data()
        if points is None or coords is None:
            return

        min_x, max_x = min_max(points[:, 0], coords[0])
        min_y, max_y = min_max(points[:, 1], coords[1])
        rect = QRectF(min_x, min_y, max_x - min_x, max_y - min_y)
        self.view_box.setRange(rect, padding=0.025)

    def update_anchors(self):
        points, labels = self.master.get_anchors()
        if points is None:
            return
        r = self.scaled_radius * np.max(np.linalg.norm(points, axis=1))
        if self.anchor_items is None:
            self.anchor_items = []
            for point, label in zip(points, labels):
                anchor = AnchorItem(line=QLineF(0, 0, *point))
                anchor._label.setToolTip(f"<b>{label}</b>")
                label = label[:MAX_LABEL_LEN - 3] + "..." if len(label) > MAX_LABEL_LEN else label
                anchor.setText(label)
                anchor.setFont(self.parameter_setter.anchor_font)

                visible = self.always_show_axes or np.linalg.norm(point) > r
                anchor.setVisible(bool(visible))
                anchor.setPen(pg.mkPen((100, 100, 100)))
                self.plot_widget.addItem(anchor)
                self.anchor_items.append(anchor)
        else:
            for anchor, point, label in zip(self.anchor_items, points, labels):
                anchor.setLine(QLineF(0, 0, *point))
                visible = self.always_show_axes or np.linalg.norm(point) > r
                anchor.setVisible(bool(visible))
                anchor.setFont(self.parameter_setter.anchor_font)

    def update_circle(self):
        super().update_circle()

        if self.always_show_axes:
            self.plot_widget.removeItem(self.circle_item)
            self.circle_item = None

        if self.circle_item is not None:
            points, _ = self.master.get_anchors()
            if points is None:
                return

            r = self.scaled_radius * np.max(np.linalg.norm(points, axis=1))
            self.circle_item.setRect(QRectF(-r, -r, 2 * r, 2 * r))
            color = self.plot_widget.palette().color(QPalette.Disabled, QPalette.Text)
            pen = pg.mkPen(color, width=1, cosmetic=True)
            self.circle_item.setPen(pen)


Placement = Enum("Placement", dict(Circular=0, LDA=1, PCA=2), type=int,
                 qualname="Placement")


class OWLinearProjection(OWAnchorProjectionWidget,
                         VizRankMixin(LinearProjectionVizRank)):
    name = __("name")
    description = __("desc")
    icon = "icons/LinearProjection.svg"
    priority = 240
    keywords = "linear projection"

    Projection_name = {Placement.Circular: __("btn.circular_placement"),
                       Placement.LDA: __("btn.linear_discriminant_analysis"),
                       Placement.PCA: __("btn.principal_component_analysis")}

    settings_version = 6

    placement = Setting(Placement.Circular)
    selected_vars = ContextSetting([])
    GRAPH_CLASS = OWLinProjGraph
    graph = SettingProvider(OWLinProjGraph)
    n_attrs_vizrank = Setting(3)

    class Error(OWAnchorProjectionWidget.Error):
        no_cont_features = Msg(__("msg.plot_requirement"))

    class Information(OWAnchorProjectionWidget.Information):
        no_lda = Msg("LDA placement is disabled due to unsuitable target.\n{}")

    def _add_controls(self):
        box = gui.vBox(self.controlArea, box=i18n.t("common.general.features"))
        self._add_controls_variables(box)
        self._add_controls_placement(box)
        super()._add_controls()
        self.gui.add_control(
            self._effects_box, gui.hSlider, __("row_hide_radius"), master=self.graph,
            value="hide_radius", minValue=0, maxValue=100, step=10,
            createLabel=False, callback=self.__radius_slider_changed
        )

    def _add_controls_variables(self, box):
        self.model_selected = VariableSelectionModel(self.selected_vars)
        variables_selection(box, self, self.model_selected)
        self.model_selected.selection_changed.connect(
            self.__model_selected_changed)
        self.btn_vizrank = self.vizrank_button(__("btn.suggest_features"))
        self.vizrankSelectionChanged.connect(self.vizrank_set_attrs)
        self.vizrankRunStateChanged.connect(self.store_vizrank_n_attrs)
        box.layout().addWidget(self.btn_vizrank)

    def _add_controls_placement(self, box):
        self.radio_placement = gui.radioButtonsInBox(
            box, self, "placement",
            btnLabels=[self.Projection_name[x] for x in Placement],
            callback=self.__placement_radio_changed
        )

    def _add_buttons(self):
        self.gui.box_zoom_select(self.buttonsArea)
        gui.auto_send(self.buttonsArea, self, "auto_commit")

    @property
    def continuous_variables(self):
        if self.data is None or self.data.domain is None:
            return []
        dom = self.data.domain
        return [v for v in chain(dom.variables, dom.metas) if v.is_continuous]

    @property
    def effective_variables(self):
        return self.selected_vars

    @property
    def effective_data(self):
        cvs = None
        if self.placement == Placement.LDA:
            cvs = self.data.domain.class_vars
        return self.data.transform(Domain(self.effective_variables, cvs))

    def vizrank_set_attrs(self, attrs):
        if not attrs:
            return
        # False positive, pylint: disable=unsupported-assignment-operation
        self.selected_vars[:] = attrs
        # Ugly, but the alternative is to have yet another signal to which
        # the view will have to connect
        self.model_selected.selection_changed.emit()

    def store_vizrank_n_attrs(self, state, data):
        if state == vizrank.RunState.Running:
            self.n_attrs_vizrank = data["n_attrs"]

    def __model_selected_changed(self):
        self.projection = None
        self._check_options()
        self.init_projection()
        self.setup_plot()
        self.commit.deferred()

    def __placement_radio_changed(self):
        self.controls.graph.hide_radius.setEnabled(
            self.placement != Placement.Circular)
        self.projection = self.projector = None
        self.init_vizrank()
        self.init_projection()
        self.setup_plot()
        self.commit.deferred()

    def __radius_slider_changed(self):
        self.graph.update_radius()

    def colors_changed(self):
        super().colors_changed()
        self.init_vizrank()

    @OWAnchorProjectionWidget.Inputs.data
    def set_data(self, data):
        super().set_data(data)
        self._check_options()
        self.init_vizrank()
        self.init_projection()

    def _check_options(self):
        buttons = self.radio_placement.buttons
        for btn in buttons:
            btn.setEnabled(True)

        problem = None
        if self.data is not None:
            if (class_var := self.data.domain.class_var) is None:
                problem = "Current data has no target variable"
            elif not class_var.is_discrete:
                problem = f"{class_var.name} is not categorical"
            elif (nclasses := len(distinct := np.unique(self.data.Y))) == 0:
                problem = f"Data has no defined values for {class_var.name}"
            elif nclasses < 3:
                vals = " and ".join(f"'{class_var.values[int(i)]}'" for i in distinct)
                problem = \
                    f"Data contains just {['one', 'two'][nclasses - 1]} distinct " \
                    f"{pl(nclasses, 'value')} ({vals}) for '{class_var.name}'; " \
                    "at least three are required."
        if problem is None:
            self.Information.no_lda.clear()
        else:
            self.Information.no_lda(problem)
            buttons[Placement.LDA].setEnabled(False)
            if self.placement == Placement.LDA:
                self.placement = Placement.Circular

        self.controls.graph.hide_radius.setEnabled(
            self.placement != Placement.Circular)

    def init_vizrank(self):
        msg = ""
        if self.data is None:
            msg = __("msg.no_data")
        elif self.attr_color is None:
            msg = __("msg.select_color_variable")
        elif self.attr_color.is_continuous and \
                self.placement == Placement.LDA:
            msg = __("msg.select_continuous_color_variable")
        elif len([v for v in self.continuous_variables
                  if v is not self.attr_color]) < 3:
            msg = __("msg.continuous_variables_lack")
        elif np.sum(np.all(np.isfinite(self.data.X), axis=1)) < 2:
            msg = __("msg.data_instances_lack")
        if not msg:
            super().init_vizrank(
                self.data, self.continuous_variables, self.attr_color,
                self.n_attrs_vizrank)
        else:
            self.disable_vizrank(msg)

    def check_data(self):
        def error(err):
            err()
            self.data = None

        super().check_data()
        if self.data is not None:
            if not len(self.continuous_variables):
                error(self.Error.no_cont_features)

    def init_attr_values(self):
        super().init_attr_values()
        self.selected_vars[:] = self.continuous_variables[:3]
        self.model_selected[:] = self.continuous_variables

    def init_projection(self):
        if self.placement == Placement.Circular:
            self.projector = CircularPlacement()
        elif self.placement == Placement.LDA:
            self.projector = LDA(solver="eigen", n_components=2)
        elif self.placement == Placement.PCA:
            self.projector = PCA(n_components=2)
            self.projector.component = 2
            self.projector.preprocessors = PCA.preprocessors + [Normalize()]

        super().init_projection()

    def get_coordinates_data(self):
        def normalized(a):
            span = np.max(a, axis=0) - np.min(a, axis=0)
            span[span == 0] = 1
            return (a - np.mean(a, axis=0)) / span

        embedding = self.get_embedding()
        if embedding is None:
            return None, None
        norm_emb = normalized(embedding[self.valid_data])
        return (norm_emb.ravel(), np.zeros(len(norm_emb), dtype=float)) \
            if embedding.shape[1] == 1 else norm_emb.T

    def _get_send_report_caption(self):
        def projection_name():
            return self.Projection_name[self.placement]

        return report.render_items_vert((
            ("Projection", projection_name()),
            ("Color", self._get_caption_var_name(self.attr_color)),
            ("Label", self._get_caption_var_name(self.attr_label)),
            ("Shape", self._get_caption_var_name(self.attr_shape)),
            ("Size", self._get_caption_var_name(self.attr_size)),
            ("Jittering", self.graph.jitter_size != 0 and
             "{} %".format(self.graph.jitter_size))))

    @classmethod
    def migrate_settings(cls, settings_, version):
        if version < 2:
            settings_["point_width"] = settings_["point_size"]
        if version < 3:
            settings_graph = {}
            settings_graph["jitter_size"] = settings_["jitter_value"]
            settings_graph["point_width"] = settings_["point_width"]
            settings_graph["alpha_value"] = settings_["alpha_value"]
            settings_graph["class_density"] = settings_["class_density"]
            settings_["graph"] = settings_graph
        if version < 4:
            if "radius" in settings_:
                settings_["graph"]["hide_radius"] = settings_["radius"]
            if "selection_indices" in settings_ and \
                    settings_["selection_indices"] is not None:
                selection = settings_["selection_indices"]
                settings_["selection"] = [(i, 1) for i, selected in
                                          enumerate(selection) if selected]
        if version < 5:
            if "placement" in settings_ and \
                    settings_["placement"] not in Placement:
                settings_["placement"] = Placement.Circular

    @classmethod
    def migrate_context(cls, context, version):
        values = context.values
        if version < 2:
            domain = context.ordered_domain
            c_domain = [t for t in context.ordered_domain if t[1] == 2]
            d_domain = [t for t in context.ordered_domain if t[1] == 1]
            for d, old_val, new_val in ((domain, "color_index", "attr_color"),
                                        (d_domain, "shape_index", "attr_shape"),
                                        (c_domain, "size_index", "attr_size")):
                index = context.values[old_val][0] - 1
                values[new_val] = (d[index][0], d[index][1] + 100) \
                    if 0 <= index < len(d) else None
        if version < 3:
            values["graph"] = {
                "attr_color": values["attr_color"],
                "attr_shape": values["attr_shape"],
                "attr_size": values["attr_size"]
            }
        if version == 3:
            values["attr_color"] = values["graph"]["attr_color"]
            values["attr_size"] = values["graph"]["attr_size"]
            values["attr_shape"] = values["graph"]["attr_shape"]
            values["attr_label"] = values["graph"]["attr_label"]
        if version < 6 and "selected_vars" in values:
            values["selected_vars"] = (values["selected_vars"], -3)

    # for backward compatibility with settings < 6, pull the enum from global
    # namespace into class
    Placement = Placement


class CircularPlacement(LinearProjector):
    def get_components(self, X, Y):
        # Return circular axes for linear projection
        n_axes = X.shape[1]
        if n_axes == 1:
            axes_angle = [0]
        elif n_axes == 2:
            axes_angle = [0, np.pi / 2]
        else:
            axes_angle = np.linspace(0, 2 * np.pi, n_axes,
                                     endpoint=False)
        return np.vstack((np.cos(axes_angle), np.sin(axes_angle)))


if __name__ == "__main__":  # pragma: no cover
    iris = Table("iris")
    WidgetPreview(OWLinearProjection).run(set_data=iris,
                                          set_subset_data=iris[::10])



import unittest

from Orange.data import Table, Domain, DiscreteVariable
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.utils.signals import Input
from Orange.widgets.utils.multi_target import check_multiple_targets_input
from Orange.widgets.widget import OWWidget


class TestMultiTargetDecorator(WidgetTest):
    class MockWidget(OWWidget):
        name = "MockWidget"

        NotCalled = object()

        class Inputs:
            data = Input("Data", Table)

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.called_with = self.NotCalled

        @Inputs.data
        @check_multiple_targets_input
        def set_data(self, obj):
            self.called_with = obj

        def pop_called_with(self):
            t = self.called_with
            self.called_with = self.NotCalled
            return t

    def setUp(self):
        self.widget = self.create_widget(self.MockWidget)
        self.data = Table("iris")

    def test_check_multiple_targets_input(self):
        class_vars = [self.data.domain.class_var,
                      DiscreteVariable("c1", values=("a", "b"))]
        domain = Domain(self.data.domain.attributes, class_vars=class_vars)
        multiple_targets_data = self.data.transform(domain)
        self.send_signal(self.widget.Inputs.data, multiple_targets_data)
        self.assertTrue(self.widget.Error.multiple_targets_data.is_shown())
        self.assertIs(self.widget.pop_called_with(), None)

        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertFalse(self.widget.Error.multiple_targets_data.is_shown())
        self.assertIs(self.widget.pop_called_with(), self.data)


if __name__ == "__main__":
    unittest.main()



from functools import wraps

from Orange.widgets.utils.messages import UnboundMsg

multiple_targets_msg = "Multiple targets are not supported."
_multiple_targets_data = UnboundMsg(multiple_targets_msg)


def check_multiple_targets_input(f):
    """
    Wrapper for widget's set_data method that checks if the input
    has multiple targets and shows an error if it does.

    :param f: widget's `set_data` method to wrap
    :return: wrapped method that handles multiple targets data inputs
    """

    @wraps(f)
    def new_f(widget, data, *args, **kwargs):
        widget.Error.add_message("multiple_targets_data",
                                 _multiple_targets_data)
        widget.Error.multiple_targets_data.clear()
        if data is not None and len(data.domain.class_vars) > 1:
            widget.Error.multiple_targets_data()
            data = None
        return f(widget, data, *args, **kwargs)

    return new_f



import unittest
import csv
import io
from AnyQt.QtCore import Qt

from AnyQt.QtWidgets import QComboBox, QWidget
from AnyQt.QtTest import QSignalSpy, QTest

from Orange.widgets.utils import textimport
from Orange.widgets.tests.base import GuiTest
from Orange.widgets.utils.textimport import TablePreview, TablePreviewModel

ColumnTypes = textimport.ColumnType

DATA1 = b",A,B,C"
DATA2 = b""
DATA3 = b"A"
DATA4 = b'A, B, C, D\n' \
        b'a, 1, 2, *\n' \
        b'b, 2, 4, *'
DATA5 = b'a\tb\n' * 1000


class OptionsWidgetTests(GuiTest):
    def test_options_widget(self):
        w = textimport.CSVOptionsWidget()
        schanged = QSignalSpy(w.optionsChanged)
        sedited = QSignalSpy(w.optionsEdited)
        w.setDialect(csv.excel())
        self.assertEqual(len(schanged), 1)
        self.assertEqual(len(sedited), 0)
        w.setSelectedEncoding("iso8859-1")

        self.assertEqual(len(schanged), 2)
        self.assertEqual(len(sedited), 0)

        d = w.dialect()
        self.assertEqual(d.delimiter, csv.excel.delimiter)
        self.assertEqual(d.doublequote, csv.excel.doublequote)
        self.assertEqual(w.encoding(), "iso8859-1")

        d = textimport.Dialect("a", "b", "c", True, True)
        w.setDialect(d)

        cb = w.findChild(QComboBox, "delimiter-combo-box")
        self.assertEqual(cb.currentIndex(),
                         textimport.CSVOptionsWidget.DelimiterOther)
        le = w.findChild(QWidget, "custom-delimiter-edit")
        self.assertEqual(le.text(), "a")

        cb = w.findChild(QWidget, "quote-edit-combo-box")
        self.assertEqual(cb.currentText(), "b")
        d1 = w.dialect()
        self.assertEqual(d.delimiter, d1.delimiter)
        self.assertEqual(d.quotechar, d1.quotechar)


class ImportWidgetTest(GuiTest):
    def test_import_widget(self):
        w = textimport.CSVImportWidget()
        w.setDialect(csv.excel())
        w.setSampleContents(io.BytesIO(DATA1))
        view = w.dataview
        model = view.model()
        self.assertEqual(model.columnCount(), 4)
        self.assertEqual(model.rowCount(), 1)
        self.assertEqual(model.canFetchMore(), False)
        w.setSampleContents(io.BytesIO(DATA2))
        model = view.model()
        self.assertEqual(model.columnCount(), 0)
        self.assertEqual(model.rowCount(), 0)
        self.assertEqual(model.canFetchMore(), False)
        w.setSampleContents(io.BytesIO(DATA4))
        model = view.model()
        self.assertEqual(model.columnCount(), 4)
        self.assertEqual(model.rowCount(), 3)

        types = {
            0: ColumnTypes.Categorical,
            1: ColumnTypes.Numeric,
            2: ColumnTypes.Text,
            3: ColumnTypes.Time,
        }
        w.setColumnTypes(types)
        self.assertEqual(w.columnTypes(), types)
        rs = w.rowStates()
        self.assertEqual(rs, {})
        w.setStateForRow(0, textimport.TablePreview.Header)
        w.setRowStates({0: textimport.TablePreview.Header})
        self.assertEqual(w.rowStates(), {0: textimport.TablePreview.Header})
        w.setStateForRow(1, textimport.TablePreview.Skipped)
        view.grab()

        w.setSampleContents(io.BytesIO(DATA5))
        model = view.model()
        self.assertEqual(model.columnCount(), 1)
        w.setDialect(csv.excel_tab())
        w.setSampleContents(io.BytesIO(DATA5))
        model = view.model()
        self.assertEqual(model.columnCount(), 2)
        self.assertTrue(model.canFetchMore())
        rows = model.rowCount()
        spy = QSignalSpy(model.rowsInserted)
        model.fetchMore()
        self.assertGreater(model.rowCount(), rows)
        self.assertEqual(len(spy), 1)

    def test_preview_view(self):
        w = TablePreview()
        model = TablePreviewModel()
        model.setPreviewStream(csv.reader(io.StringIO(DATA4.decode('utf-8'))))
        w.setModel(model)
        QTest.mouseClick(w.verticalHeader().viewport(), Qt.LeftButton)
        self.assertEqual(w.selectionBehavior(), TablePreview.SelectRows)
        QTest.mouseClick(w.horizontalHeader().viewport(), Qt.LeftButton)
        self.assertEqual(w.selectionBehavior(), TablePreview.SelectColumns)

        QTest.mouseClick(w.verticalHeader().viewport(), Qt.LeftButton)
        self.assertEqual(w.selectionBehavior(), TablePreview.SelectRows)
        QTest.mouseClick(w.viewport(), Qt.LeftButton)
        self.assertEqual(w.selectionBehavior(), TablePreview.SelectColumns)


if __name__ == "__main__":
    unittest.main(__name__)



import numpy as np
import scipy.sparse as sp
from sklearn import decomposition as skl_decomposition

import Orange.data
from Orange.data import Variable
from Orange.data.util import get_unique_names
from Orange.misc.wrapper_meta import WrapperMeta
from Orange.preprocess.score import LearnerScorer
from Orange.projection import SklProjector, DomainProjection

__all__ = ["PCA", "SparsePCA", "IncrementalPCA", "TruncatedSVD"]


class _FeatureScorerMixin(LearnerScorer):
    feature_type = Variable
    component = 0

    def score(self, data):
        model = self(data)
        return (
            np.abs(model.components_[:self.component]) if self.component
            else np.abs(model.components_),
            model.orig_domain.attributes)


class PCA(SklProjector, _FeatureScorerMixin):
    __wraps__ = skl_decomposition.PCA
    name = 'PCA'
    supports_sparse = True

    def __init__(self, n_components=None, copy=True, whiten=False,
                 svd_solver='auto', tol=0.0, iterated_power='auto',
                 random_state=None, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.params = vars()

    def fit(self, X, Y=None):
        params = self.params.copy()
        if params["n_components"] is not None:
            params["n_components"] = min(min(X.shape), params["n_components"])

        # scikit-learn doesn't support requesting the same number of PCs as
        # there are columns when the data is sparse. In this case, densify the
        # data. Since we're essentially requesting back a PC matrix of the same
        # size as the original data, we will assume the matrix is small enough
        # to densify as well
        if sp.issparse(X) and params["n_components"] == min(X.shape):
            X = X.toarray()

        # In scikit-learn==1.4.0, only the arpack solver is supported for sparse
        # data and `svd_solver="auto"` doesn't auto-resolve to this. This is
        # fixed in scikit-learn 1.5.0, but for the time being, override these
        # settings here
        if sp.issparse(X) and params["svd_solver"] == "auto":
            params["svd_solver"] = "arpack"

        proj = self.__wraps__(**params)
        proj = proj.fit(X, Y)
        return PCAModel(proj, self.domain, len(proj.components_))


class SparsePCA(SklProjector):
    __wraps__ = skl_decomposition.SparsePCA
    name = 'Sparse PCA'
    supports_sparse = False

    def __init__(self, n_components=None, alpha=1, ridge_alpha=0.01,
                 max_iter=1000, tol=1e-8, method='lars', n_jobs=1, U_init=None,
                 V_init=None, verbose=False, random_state=None, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.params = vars()

    def fit(self, X, Y=None):
        proj = self.__wraps__(**self.params)
        proj = proj.fit(X, Y)
        return PCAModel(proj, self.domain, len(proj.components_))


class PCAModel(DomainProjection, metaclass=WrapperMeta):
    var_prefix = "PC"

    def _get_var_names(self, n):
        names = [f"{self.var_prefix}{postfix}" for postfix in range(1, n + 1)]
        return get_unique_names(self.orig_domain, names)


class IncrementalPCA(SklProjector):
    __wraps__ = skl_decomposition.IncrementalPCA
    name = 'Incremental PCA'
    supports_sparse = False

    def __init__(self, n_components=None, whiten=False, copy=True,
                 batch_size=None, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.params = vars()

    def fit(self, X, Y=None):
        proj = self.__wraps__(**self.params)
        proj = proj.fit(X, Y)
        return IncrementalPCAModel(proj, self.domain, len(proj.components_))

    def partial_fit(self, data):
        return self(data)


class IncrementalPCAModel(PCAModel):
    def partial_fit(self, data):
        if isinstance(data, Orange.data.Storage):
            if data.domain != self.pre_domain:
                data = data.from_table(self.pre_domain, data)
            self.proj.partial_fit(data.X)
        else:
            self.proj.partial_fit(data)
        self.__dict__.update(self.proj.__dict__)
        return self


class TruncatedSVD(SklProjector, _FeatureScorerMixin):
    __wraps__ = skl_decomposition.TruncatedSVD
    name = 'Truncated SVD'
    supports_sparse = True

    def __init__(self, n_components=2, algorithm='randomized', n_iter=5,
                 random_state=None, tol=0.0, preprocessors=None):
        super().__init__(preprocessors=preprocessors)
        self.params = vars()

    def fit(self, X, Y=None):
        params = self.params.copy()
        # strict requirement in scikit fit_transform:
        # n_components must be < n_features
        params["n_components"] = min(min(X.shape) - 1, params["n_components"])

        proj = self.__wraps__(**params)
        proj = proj.fit(X, Y)
        return PCAModel(proj, self.domain, len(proj.components_))



"""
Widgets from Evaluate category

"""
from Orange.i18n_config import *


def __(key):
    return i18n.t("widget.evaluate.evaluate." + key)



NAME = "Evaluate"
ID = "orange.widgets.evaluate"
LABEL = __("package_label")

DESCRIPTION = __("package_desc")

BACKGROUND = "#C3F3F3"

ICON = "icons/Category-Evaluate.svg"

PRIORITY = 5
